#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2016-06-13
# @Filename: analysis_props.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-07-22 01:52:54


from __future__ import absolute_import, division, print_function

from astropy.units import Quantity, dimensionless_unscaled

from .base_quantity import QuantityMixIn


__all__ = ['AnalysisProperty']


class AnalysisProperty(Quantity, QuantityMixIn):
    """A class describing a measurement with additional information.

    Represents a quantity with an associated unit and possibly an associated
    error and mask.

    Parameters:
        value (float):
            The value of the quantity.
        unit (astropy.unit.Unit, optional):
            The unit of the quantity.
        scale (float, optional):
            The scale factor of the quantity value.
        ivar (float or None):
            The inverse variance associated with ``value``, or ``None`` if not
            defined.
        mask (int or None):
            The mask value associated with ``value``, or ``None`` if not
            defined.
        pixmask_flag (str):
            The maskbit flag to be used to convert from mask bits to labels
            (e.g., MANGA_DAPPIXMASK).

    """

    def __new__(cls, value, unit=dimensionless_unscaled, scale=1,
                ivar=None, mask=None, pixmask_flag=None, dtype=None, copy=True):

        obj = Quantity(value, unit=unit, dtype=dtype, copy=copy)
        obj = obj.view(cls)
        obj._set_unit(unit)

        obj.ivar = ivar
        obj.mask = mask
        obj.pixmask_flag = pixmask_flag

        return obj

    def __array_finalize__(self, obj):

        if obj is None:
            return

        self.ivar = getattr(obj, 'ivar', None)
        self.mask = getattr(obj, 'mask', None)

        self._set_unit(getattr(obj, 'unit', None))

    @property
    def std(self):
        """The standard deviation of the measurement."""

        return self.error



# encoding: utf-8
#
# setup.py
#

from setuptools import setup
import os

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


filedata = urlopen(
    'https://svn.sdss.org/public/repo/sdss/idlutils/trunk/data/sdss/sdssMaskbits.par')
datatowrite = filedata.read()
with open(os.path.join(os.path.dirname(__file__),
          'python/marvin/data/sdssMaskbits.par'), 'wb') as f:
    f.write(datatowrite)


setup()



#!/usr/bin/env python
# encoding: utf-8
#
# Licensed under a 3-clause BSD license.
#
# Original code from mangadap.plot.colorbar.py licensed under the following
# 3-clause BSD license.
#
# Copyright (c) 2015, SDSS-IV/MaNGA Pipeline Group
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# utils.py
#
# Created by José Sánchez-Gallego on 5 Aug 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from matplotlib import pyplot as plt


__ALL__ = ('bind_to_figure')


def bind_to_figure(ax, fig=None):
    """Copies axes to a new figure.

    This is a custom implementation of a method to copy axes from one
    matplotlib figure to another. Matplotlib does not allow this, so we create
    a new figure and copy the relevant lines and containers.

    This is all quite hacky and may stop working in future versions of
    matplotlib, but it seems to be the only way to bind axes from one figure
    to a different one.

    Current limitations include: 1) the legend is copied but its style is not
    maintained; 2) scatter plots do not maintain the marker type, markers are
    always replaced with squares.

    """

    if fig is not None:
        assert isinstance(fig, plt.Figure), 'argument must be a Figure'
        assert len(fig.axes) == 1, 'figure must have one and only one axes'
        new_ax = fig.axes[0]
    else:
        fig, new_ax = plt.subplots()

    new_ax.set_facecolor(ax.get_facecolor())

    for line in ax.lines:
        data = line.get_data()
        new_ax.plot(data[0], data[1], linestyle=line.get_linestyle(), color=line.get_color(),
                    zorder=line.zorder, label=line.get_label())

    for collection in ax.collections:
        data = collection.get_offsets()
        new_ax.scatter(data[:, 0], data[:, 1], marker='s', facecolor=collection.get_facecolors(),
                       edgecolor=collection.get_edgecolors(), s=collection.get_sizes(),
                       zorder=line.zorder, label=collection.get_label())

    for text in ax.texts:
        xx, yy = text.get_position()
        new_ax.text(xx, yy, text.get_text(), family=text.get_fontfamily(),
                    fontsize=text.get_fontsize(),
                    color=text.get_color(), ha=text.get_horizontalalignment(),
                    va=text.get_verticalalignment(), zorder=text.zorder)

    for image in ax.images:
        new_ax.imshow(image.get_array(), interpolation=image.get_interpolation())

    if ax.legend_:
        new_ax.legend()

    new_ax.grid(ax.get_xgridlines(), color=ax.get_xgridlines()[0].get_color(),
                alpha=ax.get_xgridlines()[0].get_alpha())
    new_ax.grid(ax.get_ygridlines(), color=ax.get_xgridlines()[0].get_color(),
                alpha=ax.get_xgridlines()[0].get_alpha())

    new_ax.set_xlim(ax.get_xlim())
    new_ax.set_ylim(ax.get_ylim())
    new_ax.set_xlabel(ax.get_xlabel())
    new_ax.set_ylabel(ax.get_ylabel())

    return fig



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-08-21 17:11:22
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-08 16:21:30

from __future__ import print_function, division, absolute_import
from marvin import config
from marvin.utils.datamodel.dap import datamodel
from marvin.core.exceptions import MarvinUserWarning
from marvin.utils.general import invalidArgs, isCallableWithArgs
from matplotlib.gridspec import GridSpec
from collections import defaultdict, OrderedDict
from astropy.visualization import hist as ahist
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats
import six
import pandas as pd
import itertools
import warnings

try:
    import mpl_scatter_density as msd
except ImportError as e:
    msd = None
    msderr = ('mpl-scatter-density is required to plot large results and was not found.  '
              'To use this feature, please install the python package!')


def compute_stats(data):
    ''' Compute some statistics given a data array

    Computes some basic statistics given a data array, excluding NaN values.
    Computes and returns the following Numpy statistics: mean, standard deviation,
    median, and the 10th, 25th, 75th, and 90th percentiles.

    Parameters:
        data (list|ndarray):
            A list or Numpy array of data

    Returns:
        A dictionary of statistics values

    '''
    stats = {'mean': np.nanmean(data), 'std': np.nanstd(data), 'median': np.nanmedian(data),
             'per10': np.nanpercentile(data, 10), 'per25': np.nanpercentile(data, 25),
             'per75': np.nanpercentile(data, 75), 'per90': np.nanpercentile(data, 90)}

    return stats


def _make_masked(data, mask=None):
    ''' Makes a masked array '''

    arr_data = data
    if not isinstance(data, np.ma.MaskedArray):
        # mask out NaN values if a mask not provided
        warnings.warn("Masking out NaN values!", MarvinUserWarning)
        mask = mask if mask else np.isnan(data)
        # create array
        arr_data = np.ma.MaskedArray(data, mask=mask)

    return arr_data


def _create_figure(hist=None, hist_axes_visible=None, use_density=None):
    ''' Create a generic figure and axis '''

    # use a scatter density projection or not
    projection = 'scatter_density' if use_density else None

    # check if mpl-scatter-density if installed
    if not msd:
        raise ImportError(msderr)

    # create the figure
    fig = plt.figure()
    ax_hist_x = None
    ax_hist_y = None

    # create axes with or without histogram
    if hist:
        if hist is True:
            gs = GridSpec(4, 4)
            ax_scat = fig.add_subplot(gs[1:4, 0:3], projection=projection)
            ax_hist_x = fig.add_subplot(gs[0, 0:3])
            ax_hist_y = fig.add_subplot(gs[1:4, 3])
        elif hist == 'x':
            gs = GridSpec(2, 1, height_ratios=[1, 2])
            ax_scat = fig.add_subplot(gs[1], projection=projection)
            ax_hist_x = fig.add_subplot(gs[0])
        elif hist == 'y':
            gs = GridSpec(1, 2, width_ratios=[2, 1])
            ax_scat = fig.add_subplot(gs[0], projection=projection)
            ax_hist_y = fig.add_subplot(gs[1])
    else:
        ax_scat = fig.add_subplot(1, 1, 1, projection=projection)

    # turn off histogram axes
    if ax_hist_x:
        plt.setp(ax_hist_x.get_xticklabels(), visible=hist_axes_visible)
    if ax_hist_y:
        plt.setp(ax_hist_y.get_yticklabels(), visible=hist_axes_visible)

    return fig, ax_scat, ax_hist_x, ax_hist_y


def _create_hist_title(data):
    ''' create a title for the histogram '''
    stats = compute_stats(data)
    hist_title = 'Stats: $\\mu={mean:.3f}, \\sigma={std:.3f}$'.format(**stats)
    return hist_title


def _get_dap_datamodel_property_label(quantity):
    ''' Format a DAP datamodel property string label '''
    return '{0} [{1}]'.format(quantity.to_string('latex'), quantity.unit.to_string('latex'))


def _get_axis_label(column, axis=''):
    ''' Create an axis label '''

    from marvin.utils.datamodel.query.base import QueryParameter
    from marvin.utils.datamodel.dap.base import Property

    if isinstance(column, QueryParameter):
        if hasattr(column, 'property') and column.property:
            label = _get_dap_datamodel_property_label(column.property)
        else:
            label = column.display
    elif isinstance(column, Property):
        label = _get_dap_datamodel_property_label(column)
    elif isinstance(column, six.string_types):
        label = column
    else:
        # label = '{0} axis'.format(axis).strip()
        label = ''

    return label


def _set_options():
    ''' Set some default Matplotlib options '''
    mpl.rcParams['axes.axisbelow'] = True
    mpl.rcParams['grid.color'] = 'gray'
    mpl.rcParams['grid.linestyle'] = 'dashed'
    mpl.rcParams['grid.alpha'] = 0.8


def _set_limits(column, lim=None, sigma_cutoff=50, percent_clip=1):
    ''' Set an axis limit

    Determines whether to apply percentile clipping or not if any data
    has a zscore value above the sigma_cutoff value.  Applies percentile clipping
    centered around the mean.

    Parameters:
        column:
            The array of data to get limits of
        lim (list|tuple):
            A user provided range
        sigma_cutoff (int):
            The number of sigma away from the mean to cutoff
        percent_clip (int|tuple):
            The percent to clip off the data array.  Input values are taken as percentages.
            Can either be integer value (halved for lo,hi) or a tuple specifying lo,hi values.
            Default is 1%.

    Returns:
        A list of axis range values to use

    '''
    if lim is not None:
        assert len(lim) == 2, 'range must be a list or tuple of 2'
    else:
        # get percent clips
        if isinstance(percent_clip, (list, tuple)):
            lo, hi = percent_clip
        else:
            lo = percent_clip / 2.
            hi = 100 - lo

        zscore = stats.zscore(column)
        # use percentile limits if the max zscore is > 50 sigma away from mean/stdev
        if np.max(zscore) > sigma_cutoff:
            lim = [np.percentile(column, lo), np.percentile(column, hi)]
        else:
            pass
    return lim


def _check_input_data(coldim, col, data=None):
    ''' Check the input data

    Parameters:
        coldim (str):
            Name of the dimension
        col (str|array):
            The list or array of values.  If data keyword is specified, col is a string name
        data (Pandas.DataFrame)
            A Pandas dataframe

    Returns:
        The column of data
    '''

    # check data
    assert col is not None, 'Must provide an {0} column'.format(coldim)

    if data is not None:
        assert isinstance(col, str), '{0} must be a string name if Dataframe provided'.format(coldim)
        assert isinstance(data, pd.core.frame.DataFrame), 'data must be Pandas dataframe'
        assert col in data.columns, '{0} must be a specified column name in Pandas dataframe'.format(coldim)
        col = data[col]
    else:
        assert isinstance(col, (list, np.ndarray, pd.core.series.Series)), '{0} data must be a list, Pandas Series, or Numpy array'.format(coldim)

    return col


def _format_hist_kwargs(axis, **kwargs):
    ''' Format the histogram kwargs from plot '''
    kwargs['color'] = kwargs.get('hist_color', 'lightblue')
    if axis == 'x':
        kwargs['ylabel'] = kwargs.get('xhist_label', 'Counts')
        kwargs['title'] = kwargs.get('xhist_title', None)
    elif axis == 'y':
        kwargs['ylabel'] = kwargs.get('yhist_label', 'Counts')
        kwargs['title'] = kwargs.get('yhist_title', None)
    kwargs['color'] = kwargs.get('hist_color', 'lightblue')
    kwargs['edgecolor'] = kwargs.get('edgecolors', None)
    return kwargs


def _prep_func_kwargs(func, kwargs):
    ''' Prepare the keyword arguments for the proper function input

    Checks an input dictionary against allowed keyword arguments
    for a given function.  Returns only those usable in that function.

    Parameters:
        func:
            The name of the function to check keywords against
        kwargs (dict):
            A dictionary of keyword arguments to test

    Returns:
        A new dictionary of usable keyword arguments

    '''
    invalid = invalidArgs(func, kwargs)
    new_kwargs = kwargs.copy()
    for key in invalid:
        __ = new_kwargs.pop(key)
    if isCallableWithArgs(func, new_kwargs):
        return new_kwargs
    else:
        raise MarvinUserWarning('Cannot call func {0} with current kwargs {1}. Check your inputs'.format(func, new_kwargs))


def plot(x, y, **kwargs):
    ''' Create a scatter plot given two columns of data

    Creates a Matplotlib plot using two input arrays of data.  Creates either a Matplotlib scatter
    plot, hexbin plot, or scatter density plot depending on the size of the input data.
    For data with < 1000 values, creates a scatter plot.  For data with values between
    1000 and 500,000, creates a hexbin plot.  For data with > 500,000 values, creates
    a scatter density plot.

    By default, will also create and display histograms for the x and y data.  This can be disabled
    setting the "with_hist" keyword to False, or "x", or "y" for displaying only that column.
    Accepts all the same keyword arguments as matplotlib scatter, hexbin, and hist methods.

    See `scatter-density <https://github.com/astrofrog/mpl-scatter-density>`_
    See `matplotlib.pyplot.scatter <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.scatter>`_
    See `matplotlib.pyplot.hexbin <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.hexbin>`_

    Parameters:
        x (str|list|ndarray):
            The x array of data
        y (str|list|ndarray):
            The y array of data
        data (Pandas dataframe):
            Optional Pandas Dataframe.  x, y specify string column names in the dataframe
        xmask (ndarray):
            A mask to apply to the x-array of data
        ymask (ndarray):
            A mask to apply to the y-array of data
        with_hist (bool|str):
            If True, creates the plot with both x,y histograms.  False, disables it.  If 'x' or 'y',
            only creates that histogram.  Default is True.
        hist_axes_visible (bool):
            If True, disables the x-axis ticks for each histogram.  Default is True.
        xlim (tuple):
            A tuple limited the range of the x-axis
        ylim (tuple):
            A tuple limited the range of the y-axis
        xlabel (str|Marvin column):
            The x axis label or a Marvin DataModel Property or QueryParameter to use for display
        ylabel (str|Marvin column):
            The y axis label or a Marvin DataModel Property or QueryParameter to use for display
        bins (int|tuple):
            A number or tuple specifying the number of bins to use in the histogram.  Default is 50.  An integer
            number is adopted for both x and y bins.  A tuple is used to customize per axis.
        return_figure (bool):
            If True, return the figure and axis object.  Default is True.
        kwargs (dict):
            Any other keyword arguments to be passed to `matplotlib.pyplot.scatter <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.scatter>`_
            or `matplotlib.pyplot.hist <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.hist>`_ or
            `matplotlib.pyplot.hexbin <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.hexbin>`_.

    Returns:
        A tuple of the matplotlib figure, axes, and histogram data (if returned)

    Example:
        >>> # create a scatter plot
        >>> import numpy as np
        >>> from marvin.utils.scatter import plot
        >>> x = np.random.random(100)
        >>> y = np.random.random(100)
        >>> plot(x, y)
    '''

    # check the input data
    data = kwargs.pop('data', None)
    x = _check_input_data('x', x, data=data)
    y = _check_input_data('y', y, data=data)

    # general keyword arguments
    use_datamodel = kwargs.pop('usemodel', None)
    xmask = kwargs.pop('xmask', None)
    ymask = kwargs.pop('ymask', None)
    return_figure = kwargs.pop('return_figure', True)

    # scatterplot keyword arguments
    xlim = kwargs.pop('xlim', None)
    ylim = kwargs.pop('ylim', None)
    xlabel = kwargs.pop('xlabel', None)
    ylabel = kwargs.pop('ylabel', None)
    color = kwargs.pop('color', None)
    size = kwargs.pop('size', 20)
    marker = kwargs.pop('marker', 'o')
    edgecolors = kwargs.pop('edgecolors', 'black')

    # hexbin keywords
    gridsize = kwargs.pop('gridsize', 50)

    # histogram keywords
    with_hist = kwargs.pop('with_hist', True)
    bins = kwargs.pop('bins', ['scott', 'scott'])
    hist_axes_visible = kwargs.pop('hist_axes_visible', False)

    # convert to numpy masked arrays
    x = _make_masked(x, mask=xmask)
    y = _make_masked(y, mask=ymask)
    count = len(x)
    use_density = True if count > 500000 else False

    # create figure and axes objects
    with plt.style.context('seaborn-darkgrid'):
        fig, ax_scat, ax_hist_x, ax_hist_y = _create_figure(hist=with_hist, use_density=use_density,
                                                            hist_axes_visible=hist_axes_visible)

    # create the hexbin or scatter plot
    kind = kwargs.get('kind', None)
    assert kind in ['hex', 'scatter', 'density', 'joint', None], 'plot kind must be either scatter, hex, density, or joint'
    if count > 1000 and count <= 500000:
        scat_kwargs = _prep_func_kwargs(plt.hexbin, kwargs)
        main = ax_scat.hexbin(x, y, gridsize=gridsize, mincnt=1, cmap='inferno', **scat_kwargs)
        cb = fig.colorbar(main, ax=ax_scat, label='Counts')
        #ax_scat.grid(color='gray', linestyle='dashed', alpha=0.8)
    elif count > 500000:
        # abort if mpl-scatter-density is not installed
        if not msd:
            raise ImportError(msderr)

        scat_kwargs = _prep_func_kwargs(plt.imshow, kwargs)
        main = ax_scat.scatter_density(x, y, cmap='inferno', **scat_kwargs)
        cb = fig.colorbar(main, ax=ax_scat, label='Number of points per pixel')
        ax_scat.grid(color='gray', linestyle='dashed', alpha=0.8)
    else:
        # create the scatter plot
        scat_kwargs = _prep_func_kwargs(plt.scatter, kwargs)
        main = ax_scat.scatter(x, y, c=color, s=size, marker=marker, edgecolors=edgecolors, **scat_kwargs)
        cb = None
        #ax_scat.grid(color='gray', linestyle='dashed', alpha=0.8)

    # set limits
    xlim = _set_limits(x, lim=xlim)
    ylim = _set_limits(y, lim=ylim)
    if xlim:
        ax_scat.set_xlim(xlim)
    if ylim:
        ax_scat.set_ylim(ylim)

    # set display names
    xlabel = _get_axis_label(xlabel, axis='x')
    ylabel = _get_axis_label(ylabel, axis='y')
    ax_scat.set_xlabel(xlabel)
    ax_scat.set_ylabel(ylabel)

    # set axes object
    axes = [ax_scat]

    # create histogram dictionary
    if with_hist:
        hist_data = {}
        xbin, ybin = bins if isinstance(bins, list) else (bins, bins)

    # set x-histogram
    if ax_hist_x:
        xhist_kwargs = _format_hist_kwargs('x', **kwargs)
        #xrange = ax_scat.get_xlim()
        xhist, fig, ax_hist_x = hist(x, bins=xbin, fig=fig, ax=ax_hist_x, **xhist_kwargs)
        axes.append(ax_hist_x)
        hist_data['xhist'] = xhist
        if cb is not None:
            ocb = fig.colorbar(main, ax=ax_hist_x)
            ocb.remove()

    # set y-histogram
    if ax_hist_y:
        yhist_kwargs = _format_hist_kwargs('y', **kwargs)
        yhist, fig, ax_hist_y = hist(y, bins=ybin, fig=fig, ax=ax_hist_y, orientation='horizontal',
                                     rotate_title=True, **yhist_kwargs)
        axes.append(ax_hist_y)
        hist_data['yhist'] = yhist

    if return_figure:
        output = (fig, axes, hist_data) if with_hist else (fig, axes)
    else:
        output = hist_data if with_hist else None

    return output


def hist(arr, mask=None, fig=None, ax=None, bins=None, **kwargs):
    ''' Create a histogram of an array

    Plots a histogram of an input column of data.  Input can be a list or a Numpy
    array.  Converts the input into a Numpy MaskedArray, applying the optional mask.  If no
    mask is supplied, it masks any NaN values.  This uses
    `Astropy's enhanced hist <http://docs.astropy.org/en/stable/api/astropy.visualization.hist.html#astropy.visualization.hist>`_
    function under the hood. Accepts all the same keyword arguments as matplotlib hist method.

    Parameters:
        arr (list|ndarray):
            An array of data to plot with.  Required.
        mask (ndarray):
            A mask to use on the data, applied to the data in a Numpy Masked Array.
        fig (plt.fig):
            An optional matplotlib figure object
        ax (plt.ax):
            An optional matplotlib axis object
        bins (int):
            The number of bins to use.  Default is a `scott <http://docs.astropy.org/en/stable/visualization/histogram.html>`_ binning scheme.
        xlabel (str|Marvin Column):
            The x axis label or a Marvin DataModel Property or QueryParameter to use for display
        ylabel (str):
            The y axis label
        title (str):
            The plot title
        rotate_title (bool):
            If True, moves the title text to the right y-axis during a horizontal histogram.  Default is False.
        return_figure (bool):
            If True, return the figure and axis object.  Default is True.
        kwargs (dict):
            Any other keyword arguments to be passed to `matplotlib.pyplot.hist <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.hist>`_.

    Returns:
        tuple: histogram data, matplotlib figure, and axis objects.

        The histogram data returned is a dictionary containing::

            {
                'bins': The number of bins used,
                'counts': A list of the count of objects within each bin,
                'binedges': A list of the left binedge used in defining each bin,
                'binids': An array of the same shape as input data, containing the binid of each element,
                'indices': A dictionary of a list of array indices within each bin
            }

    Example:
        >>> # histogram some random data
        >>> from marvin.utils.plot.scatter import hist
        >>> import numpy as np
        >>> x = np.random.random(100)
        >>> hist_data, fig, ax = hist(x)
    '''

    # check the input data
    data = kwargs.pop('data', None)
    arr = _check_input_data('column', arr, data=data)
    arr = _make_masked(arr, mask=mask)

    # general keywords
    xlabel = kwargs.pop('xlabel', None)
    ylabel = kwargs.pop('ylabel', 'Counts')
    title = kwargs.pop('title', None)
    rotate_title = kwargs.pop('rotate_title', False)
    return_figure = kwargs.pop('return_figure', True)

    # histogram keywords
    bins = bins if bins else 'scott'
    color = kwargs.pop('color', None)
    edgecolor = kwargs.pop('edgecolor', None)
    hrange = kwargs.pop('range', None)
    orientation = kwargs.pop('orientation', 'vertical')

    # create a figure and axis if they don't exist
    with plt.style.context('seaborn-darkgrid'):
        if fig is None and ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            fig = plt.figure()

    # set labels
    xlabel = _get_axis_label(xlabel, axis='x')
    ax.set_ylabel(ylabel) if orientation == 'vertical' else ax.set_ylabel(xlabel)
    ax.set_xlabel(xlabel) if orientation == 'vertical' else ax.set_xlabel(ylabel)

    # reset the label positions
    ax.yaxis.set_label_position('left')
    ax.xaxis.set_label_position('bottom')

    # set limits
    hrange = _set_limits(arr, lim=hrange)

    # set title
    title = title if title else _create_hist_title(arr)
    ax.set_title(title)

    if rotate_title:
        ax.set_title('')
        ax.yaxis.set_label_position('right')
        ax.yaxis.label.set_fontsize(12.0)
        ax.set_ylabel(title, rotation=270, verticalalignment='bottom')

    # create histogram
    hist_kwargs = _prep_func_kwargs(ahist, kwargs)
    counts, binedges, patches = ahist(arr[~arr.mask], bins=bins, color=color,
                                      orientation=orientation, edgecolor=edgecolor,
                                      range=hrange, ax=ax, **hist_kwargs)

    # compute a dictionary of the binids containing a list of the array indices in each bin
    binids = np.digitize(arr, binedges)
    inds = np.where(binids)[0]
    indices = defaultdict(list)
    tmp = list(map(lambda i, x: indices[x].append(i), inds, binids))

    hist_data = {'counts': counts, 'binedges': binedges, 'bins': bins,
                 'binids': binids, 'indices': indices}

    output = (hist_data, fig, ax) if return_figure else hist_data
    return output







#!/usr/bin/env python
# encoding: utf-8

# Created by Brian Cherinka on 2016-05-17 10:17:35
# Licensed under a 3-clause BSD license.

# Revision History:
#     Initial Version: 2016-05-17 10:17:35 by Brian Cherinka
#     Last Modified On: 2016-05-17 10:17:35 by Brian

from __future__ import division, print_function

import inspect

from astropy.io import fits

from marvin import config
from marvin.core.exceptions import MarvinError
from marvin.tools.cube import Cube
from marvin.utils.general.structs import FuzzyList

from .core import MarvinToolsClass


try:
    from sdss_access.path import Path
except ImportError:
    Path = None


class Plate(MarvinToolsClass, FuzzyList):
    '''A class to interface with MaNGA Plate.

    This class represents a Plate, initialised either
    from a file, a database, or remotely via the Marvin API. The class
    inherits from Python's list class, and is defined as a list of
    Cube objects.  As it inherits from list, it can do all the standard Python
    list operations.

    When instanstantiated, Marvin Plate will attempt to discover and load all the Cubes
    associated with this plate.

    Parameters:
        plate (str):
            The plate id of the Plate to load.
        plateifu (str):
            The plate-ifu of the Plate to load
        filename (str):
            The path of the file containing the data cube to load.
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See
            :doc:`Mode secision tree</mode_decision>`..
        release (str):
            The MPL/DR version of the data to use.
        nocubes (bool):
            Set this to turn off the Cube loading

    Attributes:
        cubeXXXX (object):
            The Marvin Cube object for the given ifu, e.g. cube1901 refers to the Cube for plateifu 8485-1901
        plate/plateid (int):
            The plate id for this plate
        cartid (str):
            The cart id for this plate
        designid (int):
            The design id for this plate
        ra (float):
            The RA of the plate center
        dec (float):
            The declination of the plate center
        dateobs (str):
            The date of observation for this plate
        surveymode (str):
            The survey mode for this plate
        isbright (bool):
            True if this is a bright time plate

    Return:
        plate:
            An object representing the Plate entity. The object is a list of
            Cube objects, one for each IFU cube in the Plate entity.

    Example:
        >>> from marvin.tools.plate import Plate
        >>> plate = Plate(plate=8485)
        >>> print(plate)
        >>> <Marvin Plate (plate=8485, n_cubes=17, mode='local', data_origin='db')>
        >>>
        >>> print('Cubes found in this plate: {0}'.format(len(plate)))
        >>> Cubes found in this plate: 4
        >>>
        >>> # access the plate via index to access the individual cubes
        >>> plate[0]
        >>> <Marvin Cube (plateifu='8485-12701', mode='local', data_origin='db')>
        >>>
        >>> # or by name
        >>> plate['12702']
        >>> <Marvin Cube (plateifu='8485-12702', mode='local', data_origin='db')>
        >>>
    '''

    def __init__(self, input=None, filename=None, mangaid=None, plateifu=None,
                 mode=None, data=None, release=None, plate=None,
                 download=None, nocubes=None):

        self._cubes = None
        self._plate = None
        self._pdict = None
        self.platedir = None
        self.nocubes = nocubes

        # If plateid specified, force a temp plateifu
        if plate:
            self.plateid = plate
            plateifu = '{0}-XXXX'.format(self.plateid)

        self.plateifu = plateifu

        args = [plate, plateifu]
        assert any(args), 'Enter plate or plateifu!'

        MarvinToolsClass.__init__(self, input=input, filename=filename, mangaid=mangaid,
                                  plateifu=plateifu, mode=mode, data=data, release=release,
                                  download=download)

        # sort out any plateid, plate-ifu, mangaid name snafus
        self._sortOutNames()

        # grab the plate info
        if self.data_origin == 'file':
            self._getPlateFromFile()
        elif self.data_origin == 'db':
            self._getPlateFromDB()
        elif self.data_origin == 'api':
            self._getPlateFromAPI()

        # load the plate params and init the Marvin Cubes
        self._setParams()
        if not self.nocubes:
            self._initCubes()

    def __repr__(self):
        '''Representation for Plate.'''

        return ('<Marvin Plate (plate={self.plateid!r}, n_cubes={0}, mode={self.mode!r}, '
                'data_origin={self.data_origin!r})>'.format(len(self), self=self))

    def __dir__(self):
        ''' Overriding dir for Plate '''

        # get the attributes from the class itself
        class_members = list(list(zip(*inspect.getmembers(self.__class__)))[0])
        instance_attr = list(self.__dict__.keys())
        # get the dir from FuzzyList
        listattr = ['cube{0}'.format(i.plateifu.split('-')[1]) for i in self]
        listattr.sort()

        return listattr + sorted(class_members + instance_attr)

    def __getattr__(self, value):

        if 'cube' in value:
            ifu = value.split('cube')[-1]
            plateifu = '{0}-{1}'.format(self.plate, ifu)
            return self[plateifu]

        return super(Plate, self).__getattribute__(value)

    def _getFullPath(self, **kwargs):
        """Returns the full path of the file in the tree."""
        self.filename = super(Plate, self)._getFullPath('mangaplate', drpver=self._drpver,
                                                        plate=self.plateid, **kwargs)
        self.platedir = self.filename
        self._checkFilename()
        return self.filename

    def _getPlateFromFile(self):
        ''' Initialize a Plate from a Cube/RSS File'''

        # Load file
        try:
            self._hdr = fits.getheader(self.filename, 1)
            self.plateid = int(self._hdr['PLATEID'])
        except Exception as e:
            raise MarvinError('Could not initialize via filename: {0}'
                              .format(e))
        else:
            self.data_origin = 'file'
            self._makePdict()

    def _getPlateFromDB(self):
        ''' Initialize a Plate from the DB '''
        import sqlalchemy
        from marvin import marvindb as mdb

        if not mdb.isdbconnected:
            raise MarvinError('No db connected')

        # Grab any cube for this plate
        cube = None
        try:
            cube = mdb.session.query(mdb.datadb.Cube).join(
                mdb.datadb.PipelineInfo, mdb.datadb.PipelineVersion).\
                filter(mdb.datadb.Cube.plate == self.plateid,
                       mdb.datadb.PipelineVersion.version == self._drpver).first()
        except sqlalchemy.orm.exc.NoResultFound as ee:
            raise MarvinError('Could not retrieve Cube for plate {0}: '
                              'No Results Found: {1}'
                              .format(self.plateid, ee))

        except Exception as ee:
            raise MarvinError('Could not retrieve Cube for plate {0}: '
                              'Unknown exception: {1}'
                              .format(self.plateid, ee))
        else:
            # no cube
            if not cube:
                raise MarvinError('No cube found in db for plate {0}, drpver {1}'
                                  .format(self.plateid, self._drpver))
            # cube but no plateclass
            try:
                self._plate = cube.plateclass
            except AttributeError as ee:
                raise MarvinError('AttributeError: cube has no plateclass for plate {0}: {1}'
                                  .format(self.plateid, ee))
            else:
                self._hdr = self._plate._hdr
                self._pdict = self._plate.__dict__
                self.data_origin = 'db'

        if not self._plate:
            raise MarvinError('Could not retrieve Plate for id {0}'.format(self.plateid))

    def _getPlateFromAPI(self):
        ''' Initialize a Plate using the API '''

        # Checks that the Plate exists.
        routeparams = {'plateid': self.plateid}
        url = config.urlmap['api']['getPlate']['url'].format(**routeparams)

        # Make the API call
        response = self._toolInteraction(url)
        data = response.getData()
        self._hdr = data['header']
        self.data_origin = 'api'
        self._makePdict()

    def _initCubes(self):
        ''' Initialize a list of Marvin Cube objects '''

        _cubes = [None]
        if self.data_origin == 'file':
            sdss_path = Path(release=self.release)
            if self.dir3d == 'stack':
                cubes = sdss_path.expand('mangacube', drpver=self._drpver,
                                         plate=self.plateid, ifu='*', wave='LOG')
            else:
                cubes = sdss_path.expand('mangamastar', drpver=self._drpver,
                                         plate=self.plateid, ifu='*', wave='LOG')
            _cubes = [Cube(filename=cube, mode=self.mode, release=self.release) for cube in cubes]

        elif self.data_origin == 'db':
            _cubes = [Cube(plateifu=cube.plateifu, mode=self.mode, release=self.release)
                      for cube in self._plate.cubes]

        elif self.data_origin == 'api':
            routeparams = {'plateid': self.plateid}
            url = config.urlmap['api']['getPlateCubes']['url'].format(**routeparams)

            # Make the API call
            response = self._toolInteraction(url)
            data = response.getData()
            plateifus = data['plateifus']
            _cubes = [Cube(plateifu=pifu, mode=self.mode, release=self.release) for pifu in plateifus]

        FuzzyList.__init__(self, _cubes)
        self.mapper = (lambda e: e.plateifu)

    def _setParams(self):
        ''' Set the plate parameters '''
        self.ra = self._pdict.get('ra', None)
        self.dec = self._pdict.get('dec', None)
        self.designid = self._pdict.get('designid', None)
        self.cartid = self._pdict.get('cartid', None)
        self.dateobs = self._pdict.get('dateobs', None)
        self.platetype = self._pdict.get('platetype', None)
        self.surveymode = self._pdict.get('surveymode', None)
        self.isbright = self._pdict.get('isbright', None)
        self.dir3d = self._pdict.get('dir3d', None)
        self.plateid = int(self.plateid)

    def _makePdict(self):
        ''' Make the necessary plate dictionary '''
        self._pdict = {}
        self._pdict['ra'] = self._hdr.get('CENRA', None)
        self._pdict['dec'] = self._hdr.get('CENDEC', None)
        self._pdict['designid'] = self._hdr.get('DESIGNID', None)
        self._pdict['cartid'] = self._hdr.get('CARTID', None)
        self._pdict['dateobs'] = self._hdr.get('DATE-OBS', None)
        self._pdict['platetype'] = self._hdr.get('PLATETYP', None)
        self._pdict['surveymode'] = self._hdr.get('SRVYMODE', None)
        self._pdict['isbright'] = 'APOGEE' in self._pdict['surveymode']
        self._pdict['dir3d'] = 'mastar' if self._pdict['isbright'] else 'stack'

        self._pdict['ra'] = float(self._pdict['ra'])
        self._pdict['dec'] = float(self._pdict['dec'])
        self._pdict['designid'] = float(self._pdict['designid'])

    def _sortOutNames(self):
        ''' Sort out any name issues with plateid, plateifu, mangaid inputs '''

        if self.plateifu and 'XXX' not in self.plateifu:
            plate, ifu = self.plateifu.split('-')
            self.plateid = int(plate)

    def _checkFilename(self):
        ''' Checks the filename for a proper FITS file '''

        # if filename is not FITS, then try to load one
        if 'fits' not in self.filename.lower():

            if not Path:
                raise MarvinError('sdss_access is not installed')
            else:
                # is_public = 'DR' in self._release
                # path_release = self._release.lower() if is_public else None
                sdss_path = Path(release=self._release)

            # try a cube
            full = sdss_path.full('mangacube', drpver=self._drpver, plate=self.plateid, ifu='*', wave='LOG')
            cubeexists = sdss_path.any('', full=full)
            if cubeexists:
                file = sdss_path.one('', full=full)
            else:
                # try an rss
                full = sdss_path.full('mangarss', drpver=self._drpver, plate=self.plateid, ifu='*', wave='LOG')
                rssexists = sdss_path.any('', full=full)
                if rssexists:
                    file = sdss_path.one('', full=full)
                else:
                    file = None
            # load the file
            if file:
                self.filename = file
            else:
                self.filename = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-11-01
# @Filename: general.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2019-08-29 15:58:00


from __future__ import absolute_import, division, print_function

import collections
import contextlib
import inspect
import os
import re
import sys
import warnings
from builtins import range
from collections import OrderedDict
from functools import wraps
from pkg_resources import parse_version

import matplotlib.pyplot as plt
import numpy as np
import PIL
from astropy import table, wcs
from astropy.units.quantity import Quantity
from brain.core.exceptions import BrainError
from flask_jwt_extended import get_jwt_identity
from scipy.interpolate import griddata

import marvin
from marvin import log
from marvin.core.exceptions import MarvinError, MarvinUserWarning


try:
    from sdss_access import Access, AccessError
except ImportError:
    Access = None

try:
    from sdss_access.path import Path
except ImportError:
    Path = None

try:
    import pympler.summary
    import pympler.muppy
    import psutil
except ImportError:
    pympler = None
    psutil = None


# General utilities
__all__ = ('convertCoords', 'parseIdentifier', 'mangaid2plateifu', 'findClosestVector',
           'getWCSFromPng', 'convertImgCoords', 'getSpaxelXY',
           'downloadList', 'getSpaxel', 'get_drpall_row', 'getDefaultMapPath',
           'getDapRedux', 'get_nsa_data', '_check_file_parameters',
           'invalidArgs', 'missingArgs', 'getRequiredArgs', 'getKeywordArgs',
           'isCallableWithArgs', 'map_bins_to_column', '_sort_dir', 'get_drpall_path',
           'get_dapall_path', 'temp_setattr', 'map_dapall', 'turn_off_ion', 'memory_usage',
           'validate_jwt', 'target_status', 'target_is_observed', 'target_is_mastar',
           'get_plates', 'get_manga_image', 'check_versions', 'get_drpall_table',
           'get_dapall_table', 'get_drpall_file', 'get_dapall_file')

drpTable = {}
dapTable = {}


def validate_jwt(f):
    ''' Decorator to validate a JWT and User '''

    @wraps(f)
    def wrapper(*args, **kwargs):
        current_user = get_jwt_identity()

        if not current_user:
            raise MarvinError('Invalid user from API token!')
        else:
            marvin.config.access = 'collab'
        return f(*args, **kwargs)
    return wrapper


def getSpaxel(cube=True, maps=True, modelcube=True,
              x=None, y=None, ra=None, dec=None, xyorig=None, **kwargs):
    """Returns the |spaxel| matching certain coordinates.

    The coordinates of the spaxel to return can be input as ``x, y`` pixels
    relative to``xyorig`` in the cube, or as ``ra, dec`` celestial
    coordinates.

    This function is intended to be called by
    :func:`~marvin.tools.cube.Cube.getSpaxel` or
    :func:`~marvin.tools.maps.Maps.getSpaxel`, and provides shared code for
    both of them.

    Parameters:
        cube (:class:`~marvin.tools.cube.Cube` or None or bool)
            A :class:`~marvin.tools.cube.Cube` object with the DRP cube
            data from which the spaxel spectrum will be extracted. If None,
            the |spaxel| object(s) returned won't contain spectral information.
        maps (:class:`~marvin.tools.maps.Maps` or None or bool)
            As ``cube`` but for the :class:`~marvin.tools.maps.Maps`
            object representing the DAP maps entity. If None, the |spaxel|
            will be returned without DAP information.
        modelcube (:class:`~marvin.tools.modelcube.ModelCube` or None or bool)
            As ``cube`` but for the :class:`~marvin.tools.modelcube.ModelCube`
            object representing the DAP modelcube entity. If None, the |spaxel|
            will be returned without model information.
        x,y (int or array):
            The spaxel coordinates relative to ``xyorig``. If ``x`` is an
            array of coordinates, the size of ``x`` must much that of
            ``y``.
        ra,dec (float or array):
            The coordinates of the spaxel to return. The closest spaxel to
            those coordinates will be returned. If ``ra`` is an array of
            coordinates, the size of ``ra`` must much that of ``dec``.
        xyorig ({'center', 'lower'}):
            The reference point from which ``x`` and ``y`` are measured.
            Valid values are ``'center'``, for the centre of the
            spatial dimensions of the cube, or ``'lower'`` for the
            lower-left corner. This keyword is ignored if ``ra`` and
            ``dec`` are defined. ``xyorig`` defaults to
            ``marvin.config.xyorig.``
        kwargs (dict):
            Arguments to be passed to `~marvin.tools.spaxel.SpaxelBase`.

    Returns:
        spaxels (list):
            The |spaxel| objects for this cube/maps corresponding to the input
            coordinates. The length of the list is equal to the number
            of input coordinates.

    .. |spaxel| replace:: :class:`~marvin.tools.spaxel.Spaxel`

    """

    # TODO: for now let's put these imports here, but we should fix the
    # circular imports soon.
    import marvin.tools.cube
    import marvin.tools.maps
    import marvin.tools.modelcube
    import marvin.tools.spaxel

    # Checks that the cube and maps data are correct
    assert cube or maps or modelcube, \
        'Either cube, maps, or modelcube needs to be specified.'

    assert isinstance(cube, (marvin.tools.cube.Cube, bool)), \
        'cube is not an instance of Cube or a boolean'

    assert isinstance(maps, (marvin.tools.maps.Maps, bool)), \
        'maps is not an instance of Maps or a boolean'

    assert isinstance(modelcube, (marvin.tools.modelcube.ModelCube, bool)), \
        'modelcube is not an instance of ModelCube or a boolean'

    # Checks that we have the correct set of inputs.
    if x is not None or y is not None:
        assert ra is None and dec is None, 'Either use (x, y) or (ra, dec)'
        assert x is not None and y is not None, 'Specify both x and y'

        inputMode = 'pix'
        isScalar = np.isscalar(x)
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        coords = np.array([x, y], np.float).T

    elif ra is not None or dec is not None:
        assert x is None and y is None, 'Either use (x, y) or (ra, dec)'
        assert ra is not None and dec is not None, 'Specify both ra and dec'

        inputMode = 'sky'
        isScalar = np.isscalar(ra)
        ra = np.atleast_1d(ra)
        dec = np.atleast_1d(dec)
        coords = np.array([ra, dec], np.float).T

    else:
        raise ValueError('You need to specify either (x, y) or (ra, dec)')

    if not xyorig:
        xyorig = marvin.config.xyorig

    if isinstance(maps, marvin.tools.maps.Maps):
        ww = maps.wcs if inputMode == 'sky' else None
        cube_shape = maps._shape
    elif isinstance(cube, marvin.tools.cube.Cube):
        ww = cube.wcs if inputMode == 'sky' else None
        cube_shape = cube._shape
    elif isinstance(modelcube, marvin.tools.modelcube.ModelCube):
        ww = modelcube.wcs if inputMode == 'sky' else None
        cube_shape = modelcube._shape

    iCube, jCube = zip(convertCoords(coords, wcs=ww, shape=cube_shape,
                                     mode=inputMode, xyorig=xyorig).T)

    _spaxels = []
    for ii in range(len(iCube[0])):
        _spaxels.append(
            marvin.tools.spaxel.Spaxel(jCube[0][ii], iCube[0][ii],
                                       cube=cube, maps=maps, modelcube=modelcube, **kwargs))

    if len(_spaxels) == 1 and isScalar:
        return _spaxels[0]
    else:
        return _spaxels


def convertCoords(coords, mode='sky', wcs=None, xyorig='center', shape=None):
    """Convert input coordinates to array indices.

    Converts input positions in x, y or RA, Dec coordinates to array indices
    (in Numpy style) or spaxel extraction. In case of pixel coordinates, the
    origin of reference (either the center of the cube or the lower left
    corner) can be specified via ``xyorig``.

    If ``shape`` is defined (mandatory for ``mode='pix'``, optional for
    ``mode='sky'``) and one or more of the resulting indices are outside the
    size of the input shape, an error is raised.

    This functions is mostly intended for internal use.

    Parameters:
        coords (array):
            The input coordinates, as an array of shape Nx2.
        mode ({'sky', 'pix'}:
            The type of input coordinates, either `'sky'` for celestial
            coordinates (in the format defined in the WCS header information),
            or `'pix'` for pixel coordinates.
        wcs (None or ``astropy.wcs.WCS`` object):
            If ``mode='sky'``, the WCS solution from which the cube coordinates
            can be derived.
        xyorig (str):
            If ``mode='pix'``, the reference point from which the coordinates
            are measured. Valid values are ``'center'``, for the centre of the
            spatial dimensions of the cube, or ``'lower'`` for the lower-left
            corner.
        shape (None or array):
            If ``mode='pix'``, the shape of the spatial dimensions of the cube,
            so that the central position can be calculated.

    Returns:
        result (Numpy array):
            An array with the same shape as ``coords``, containing the cube
            index positions for the input coordinates, in Numpy style (i.e.,
            the first element being the row and the second the column).

    """

    coords = np.atleast_2d(coords)
    assert coords.shape[1] == 2, 'coordinates must be an array Nx2'

    if mode == 'sky':
        assert wcs, 'if mode==sky, wcs must be defined.'
        coordsSpec = np.ones((coords.shape[0], 3), np.float32)
        coordsSpec[:, :-1] = coords
        cubeCoords = wcs.wcs_world2pix(coordsSpec, 0)
        cubeCoords = np.fliplr(np.array(np.round(cubeCoords[:, :-1]), np.int))

    elif mode in ['pix', 'pixel']:
        assert xyorig, 'if mode==pix, xyorig must be defined.'
        x = coords[:, 0]
        y = coords[:, 1]

        assert shape is not None, 'if mode==pix, shape must be defined.'
        shape = np.atleast_1d(shape)

        if xyorig == 'center':
            yMid, xMid = shape / 2.
            xCube = np.round(xMid + x)
            yCube = np.round(yMid + y)
        elif xyorig == 'lower':
            xCube = np.round(x)
            yCube = np.round(y)
        else:
            raise ValueError('xyorig must be center or lower.')

        cubeCoords = np.array([yCube, xCube], np.int).T

    else:
        raise ValueError('mode must be pix or sky.')

    if shape is not None:
        if ((cubeCoords < 0).any() or
                (cubeCoords[:, 0] > (shape[0] - 1)).any() or
                (cubeCoords[:, 1] > (shape[1] - 1)).any()):
            raise MarvinError('some indices are out of limits.'
                              '``xyorig`` is currently set to "{0}". '
                              'Try setting ``xyorig`` to "{1}".'
                              .format(xyorig, 'center' if xyorig == 'lower' else 'lower'))

    return cubeCoords


def mangaid2plateifu(mangaid, mode='auto', drpall=None, drpver=None):
    """Return the plate-ifu for a certain mangaid.

    Uses either the DB or the drpall file to determine the plate-ifu for
    a mangaid. If more than one plate-ifu are available for a certain ifu,
    and ``mode='drpall'``, the one with the higher SN2 (calculated as the sum
    of redSN2 and blueSN2) will be used. If ``mode='db'``, the most recent one
    will be used.

    Parameters:
        mangaid (str):
            The mangaid for which the plate-ifu will be returned.
        mode ({'auto', 'drpall', 'db', 'remote'}):
            If `'drpall'` or ``'db'``, the  drpall file or the local database,
            respectively, will be used. If ``'remote'``, a request to the API
            will be issued. If ``'auto'``, the local modes will be tried before
            the remote mode.
        drpall (str or None):
            The path to the drpall file to use. If None, the file in
            ``config.drpall`` will be used.
        drpver (str or None):
            The DRP version to use. If None, the one in ``config.drpver`` will
            be used. If ``drpall`` is defined, this value is ignored.

    Returns:
        plateifu (str):
            The plate-ifu string for the input ``mangaid``.

    """

    from marvin import config, marvindb
    from marvin.api.api import Interaction

    # The modes and order over which the auto mode will loop.
    autoModes = ['db', 'drpall', 'remote']

    assert mode in autoModes + ['auto'], 'mode={0} is not valid'.format(mode)

    config_drpver, __ = config.lookUpVersions()
    drpver = drpver if drpver else config_drpver
    drpall = drpall if drpall else config._getDrpAllPath(drpver=drpver)

    if mode == 'drpall':

        # Get the drpall table from cache or fresh
        drpall_table = get_drpall_table(drpver=drpver, drpall=drpall)

        mangaids = np.array([mm.strip() for mm in drpall_table['mangaid']])

        plateifus = drpall_table[np.where(mangaids == mangaid)]

        if len(plateifus) > 1:
            warnings.warn('more than one plate-ifu found for mangaid={0}. '
                          'Using the one with the highest SN2.'.format(mangaid),
                          MarvinUserWarning)
            plateifus = plateifus[
                [np.argmax(plateifus['bluesn2'] + plateifus['redsn2'])]]

        if len(plateifus) == 0:
            raise ValueError('no plate-ifus found for mangaid={0}'.format(mangaid))

        return plateifus['plateifu'][0]

    elif mode == 'db':

        if not marvindb.isdbconnected:
            raise MarvinError('no DB connection found')

        if not drpver:
            raise MarvinError('drpver not set.')

        cubes = marvindb.session.query(marvindb.datadb.Cube).join(
            marvindb.datadb.PipelineInfo, marvindb.datadb.PipelineVersion).filter(
                marvindb.datadb.Cube.mangaid == mangaid,
                marvindb.datadb.PipelineVersion.version == drpver).use_cache().all()

        if len(cubes) == 0:
            raise ValueError('no plate-ifus found for mangaid={0}'.format(mangaid))
        elif len(cubes) > 1:
            warnings.warn('more than one plate-ifu found for mangaid={0}. '
                          'Using a the one with the higest SN2'.format(mangaid),
                          MarvinUserWarning)
            total_sn2 = [float(cube.header['BLUESN2']) + float(cube.header['REDSN2'])
                         for cube in cubes]
            cube = cubes[np.argmax(total_sn2)]
        else:
            cube = cubes[0]

        return '{0}-{1}'.format(cube.plate, cube.ifu.name)

    elif mode == 'remote':

        try:
            url = marvin.config.urlmap['api']['mangaid2plateifu']['url']
            response = Interaction(url.format(mangaid=mangaid))
        except MarvinError as e:
            raise MarvinError('API call to mangaid2plateifu failed: {0}'.format(e))
        else:
            plateifu = response.getData(astype=str)

        if not plateifu:
            if 'error' in response.results and response.results['error']:
                raise MarvinError(response.results['error'])
            else:
                raise MarvinError('API call to mangaid2plateifu failed with error unknown.')

        return plateifu

    elif mode == 'auto':

        for mm in autoModes:
            try:
                plateifu = mangaid2plateifu(mangaid, mode=mm, drpver=drpver, drpall=drpall)
                return plateifu
            except:
                continue

        raise MarvinError(
            'mangaid2plateifu was not able to find a plate-ifu for '
            'mangaid={0} either local or remotely.'.format(mangaid))


def findClosestVector(point, arr_shape=None, pixel_shape=None, xyorig=None):
    """Find the closest array coordinates from pixel coordinates.

    Find the closest vector of array coordinates (x, y) from an input
    vector of pixel coordinates (x, y).

    Parameters:
        point : tuple
            Original point of interest in pixel units, order of (x,y)
        arr_shape : tuple
            Shape of data array in (x,y) order
        pixel_shape : tuple
            Shape of image in pixels in (x,y) order
        xyorig : str
            Indicates the origin point of coordinates.  Set to
            "relative" switches to an array coordinate system relative
            to galaxy center.  Default is absolute array coordinates
            (x=0, y=0) = upper left corner

    Returns:
        minind : tuple
            A tuple of array coordinates in x, y order
    """

    # set as numpy arrays
    arr_shape = np.array(arr_shape, dtype=int)
    pixel_shape = np.array(pixel_shape, dtype=int)

    # compute midpoints
    xmid, ymid = arr_shape / 2
    xpixmid, ypixmid = pixel_shape / 2

    # default absolute array coordinates
    xcoords = np.array([0, arr_shape[0]], dtype=int)
    ycoords = np.array([0, arr_shape[1]], dtype=int)

    # split x,y coords and pixel coords
    x1, x2 = xcoords
    y1, y2 = ycoords
    xpix, ypix = pixel_shape

    # build interpolates between array coordinates and pixel coordinates
    points = [[x1, y1], [x1, y2], [xmid, ymid], [x2, y1], [x2, y2]]
    values = [[0, ypix], [0, 0], [xpixmid, ypixmid], [xpix, ypix], [xpix, 0]]  # full image
    # values = [[xpixmid-xmid, ypixmid+ymid], [xpixmid-xmid, ypixmid-ymid], [xpixmid, ypixmid], [xpixmid+xmid, ypixmid+ymid], [xpixmid+xmid, ypixmid-ymid]]  # pixels based on arr_shape
    #values = [[xpixmid-x2, ypixmid+y2], [xpixmid-x2, ypixmid-y2], [xpixmid, ypixmid], [xpixmid+x2, ypixmid+y2], [xpixmid+x2, ypixmid-y2]]  # pixels based on arr_shape

    # make 2d array of array indices in absolute or (our) relative coordindates
    arrinds = np.mgrid[x1:x2, y1:y2].swapaxes(0, 2).swapaxes(0, 1)
    # interpolate a new 2d pixel coordinate array
    final = griddata(points, values, arrinds)

    # find minimum array vector closest to input coordinate point
    diff = np.abs(point - final)
    prod = diff[:, :, 0] * diff[:, :, 1]
    minind = np.unravel_index(prod.argmin(), arr_shape)

    # toggle relative array coordinates
    if xyorig in ['relative', 'center']:
        minind = np.array(minind, dtype=int)
        xmin = minind[0] - xmid
        ymin = ymid - minind[1]
        minind = (xmin, ymin)

    return minind


def getWCSFromPng(filename=None, image=None):
    """Extract any WCS info from the metadata of a PNG image.

    Extracts the WCS metadata info from the PNG optical
    image of the galaxy using PIL (Python Imaging Library).
    Converts it to an Astropy WCS object.

    Parameters:
        image (object):
            An existing PIL image object
        filename (str):
            The full path to the image

    Returns:
        pngwcs (WCS):
            an Astropy WCS object
    """

    assert any([image, filename]), 'Must provide either a PIL image object, or the full image filepath'

    pngwcs = None

    if filename and not image:
        try:
            image = PIL.Image.open(filename)
        except Exception as e:
            raise MarvinError('Cannot open image {0}: {1}'.format(filename, e))
        else:
            # Close the image
            image.close()

    # get metadata
    meta = image.info if image else None

    # parse the image metadata
    mywcs = {}
    if meta and 'WCSAXES' in meta.keys():
        for key, val in meta.items():
            try:
                val = float(val)
            except Exception as e:
                pass
            mywcs.update({key: val})

        tmp = mywcs.pop('WCSAXES')

    # Construct Astropy WCS
    if mywcs:
        pngwcs = wcs.WCS(mywcs)

    return pngwcs


def convertImgCoords(coords, image, to_pix=None, to_radec=None):
    """Transform the WCS info in an image.

    Convert image pixel coordinates to RA/Dec based on
    PNG image metadata or vice_versa

    Parameters:
        coords (tuple):
            The input coordindates to transform
        image (str):
            The full path to the image
        to_pix (bool):
            Set to convert to pixel coordinates
        to_radec (bool):
            Set to convert to RA/Dec coordinates

    Returns:
        newcoords (tuple):
            Tuple of either (x, y) pixel coordinates
            or (RA, Dec) coordinates
    """

    try:
        wcs = getWCSFromPng(image)
    except Exception as e:
        raise MarvinError('Cannot get wcs info from image {0}: {1}'.format(image, e))

    if to_radec:
        try:
            newcoords = wcs.all_pix2world([coords], 1)[0]
        except AttributeError as e:
            raise MarvinError('Cannot convert coords to RA/Dec.  No wcs! {0}'.format(e))
    if to_pix:
        try:
            newcoords = wcs.all_world2pix([coords], 1)[0]
        except AttributeError as e:
            raise MarvinError('Cannot convert coords to image pixels.  No wcs! {0}'.format(e))
    return newcoords


def parseIdentifier(galid):
    """Determine if a string input is a plate, plateifu, or manga-id.

    Parses a string object id and determines whether it is a
    plate ID, a plate-IFU, or MaNGA-ID designation.

    Parameters:
        galid (str):
            The string of an id name to parse

    Returns:
        idtype (str):
            String indicating either plate, plateifu, mangaid, or None
    """

    galid = str(galid)
    hasdash = '-' in galid

    if hasdash:
        galidsplit = galid.split('-')

        if int(galidsplit[0]) > 6500:
            idtype = 'plateifu'
        else:
            idtype = 'mangaid'
    else:
        # check for plate
        if galid.isdigit():
            if len(galid) > 3:
                idtype = 'plate'
            else:
                idtype = None
        else:
            idtype = None

    return idtype


def getSpaxelXY(cube, plateifu, x, y):
    """Get a spaxel from a cube in the DB.

    This function is mostly intended for internal use.

    Parameters:
        cube (SQLAlchemy object):
            The SQLAlchemy object representing the cube from which to extract
            the spaxel.
        plateifu (str):
            The corresponding plateifu of ``cube``.
        x,y (int):
            The coordinates of the spaxel in the database.

    Returns:
        spaxel (SQLAlchemy object):
            The SQLAlchemy spaxel with coordinates ``(x, y)``.

    """

    import sqlalchemy

    mdb = marvin.marvindb

    try:
        spaxel = mdb.session.query(mdb.datadb.Spaxel).filter_by(cube=cube, x=x, y=y).use_cache().one()
    except sqlalchemy.orm.exc.NoResultFound as e:
        raise MarvinError('Could not retrieve spaxel for plate-ifu {0} at position {1},{2}: No Results Found: {3}'.format(plateifu, x, y, e))
    except Exception as e:
        raise MarvinError('Could not retrieve cube for plate-ifu {0} at position {1},{2}: Unknown exception: {3}'.format(plateifu, x, y, e))

    return spaxel


def getDapRedux(release=None):
    """Retrieve SAS url link to the DAP redux directory.

    Parameters:
        release (str):
            The release version of the data to download.
            Defaults to Marvin config.release.

    Returns:
        dapredux (str):
            The full redux path to the DAP MAPS
    """

    if not Path:
        raise MarvinError('sdss_access is not installed')
    else:
        # is_public = 'DR' in release
        # path_release = release.lower() if is_public else None
        sdss_path = Path(release=release)

    release = release or marvin.config.release
    drpver, dapver = marvin.config.lookUpVersions(release=release)
    ## hack a url version of MANGA_SPECTRO_ANALYSIS
    #dapdefault = sdss_path.dir('mangadefault', drpver=drpver, dapver=dapver, plate=None, ifu=None)
    #dappath = dapdefault.rsplit('/', 2)[0]
    dappath = os.path.join(os.getenv("MANGA_SPECTRO_ANALYSIS"), drpver, dapver)
    dapredux = sdss_path.url('', full=dappath)
    return dapredux


def getDefaultMapPath(**kwargs):
    """Retrieve the default Maps path.

    Uses sdss_access Path to generate a url download link to the
    default MAPS file for a given MPL.

    Parameters:
        release (str):
            The release version of the data to download.
            Defaults to Marvin config.release.
        plate (int):
            The plate id
        ifu (int):
            The ifu number
        mode (str):
            The bintype of the default file to grab, i.e. MAPS or LOGCUBE. Defaults to MAPS
        daptype (str):
            The daptype of the default map to grab.  Defaults to SPX-GAU-MILESHC

    Returns:
        maplink (str):
            The sas url to download the default maps file
    """

    # Get kwargs
    release = kwargs.get('release', marvin.config.release)
    drpver, dapver = marvin.config.lookUpVersions(release=release)
    plate = kwargs.get('plate', None)
    ifu = kwargs.get('ifu', None)
    daptype = kwargs.get('daptype', 'SPX-GAU-MILESHC')
    mode = kwargs.get('mode', 'MAPS')
    assert mode in ['MAPS', 'LOGCUBE'], 'mode can either be MAPS or LOGCUBE'

    # get sdss_access Path
    if not Path:
        raise MarvinError('sdss_access is not installed')
    else:
        # is_public = 'DR' in release
        # path_release = release.lower() if is_public else None
        sdss_path = Path(release=release)

    # get the sdss_path name by MPL
    # TODO: this is likely to break in future MPL/DRs. Just a heads up.
    if '4' in release:
        name = 'mangadefault'
    else:
        name = 'mangadap'

    # construct the url link to default maps file
    maplink = sdss_path.url(name, drpver=drpver, dapver=dapver, mpl=release,
                            plate=plate, ifu=ifu, daptype=daptype, mode=mode)
    return maplink


def downloadList(inputlist, dltype='cube', **kwargs):
    """Download a list of MaNGA objects.

    Uses sdss_access to download a list of objects
    via rsync.  Places them in your local sas path mimicing
    the Utah SAS.

    i.e. $SAS_BASE_DIR/mangawork/manga/spectro/redux

    Can download cubes, rss files, maps, modelcubes, mastar cubes,
    png images, default maps, or the entire plate directory.
    dltype=`dap` is a special keyword that downloads all DAP files.  It sets binmode
    and daptype to '*'

    Parameters:
        inputlist (list):
            Required.  A list of objects to download.  Must be a list of plate IDs,
            plate-IFUs, or manga-ids
        dltype ({'cube', 'maps', 'modelcube', 'dap', image', 'rss', 'mastar', 'default', 'plate'}):
            Indicated type of object to download.  Can be any of
            plate, cube, image, mastar, rss, map, modelcube, or default (default map).
            If not specified, the dltype defaults to cube.
        release (str):
            The MPL/DR version of the data to download.
            Defaults to Marvin config.release.
        bintype (str):
            The bin type of the DAP maps to download. Defaults to *
        binmode (str):
            The bin mode of the DAP maps to download. Defaults to *
        n (int):
            The plan id number [1-12] of the DAP maps to download. Defaults to *
        daptype (str):
            The daptype of the default map to grab.  Defaults to *
        dir3d (str):
            The directory where the images are located.  Either 'stack' or 'mastar'. Defaults to *
        verbose (bool):
            Turns on verbosity during rsync
        limit (int):
            A limit to the number of items to download
        test (bool):
            If True, tests the download path construction but does not download

    Returns:
        If test=True, returns the list of full filepaths that will be downloaded
    """

    assert isinstance(inputlist, (list, np.ndarray)), 'inputlist must be a list or numpy array'

    # Get some possible keywords
    # Necessary rsync variables:
    #   drpver, plate, ifu, dir3d, [mpl, dapver, bintype, n, mode]
    verbose = kwargs.get('verbose', None)
    as_url = kwargs.get('as_url', None)
    release = kwargs.get('release', marvin.config.release)
    drpver, dapver = marvin.config.lookUpVersions(release=release)
    bintype = kwargs.get('bintype', '*')
    binmode = kwargs.get('binmode', None)
    daptype = kwargs.get('daptype', '*')
    dir3d = kwargs.get('dir3d', '*')
    n = kwargs.get('n', '*')
    limit = kwargs.get('limit', None)
    test = kwargs.get('test', None)
    wave = 'LOG'

    # check for sdss_access
    if not Access:
        raise MarvinError('sdss_access not installed.')

    # Assert correct dltype
    dltype = 'cube' if not dltype else dltype
    assert dltype in ['plate', 'cube', 'mastar', 'modelcube', 'dap', 'rss', 'maps', 'image',
                      'default'], ('dltype must be one of plate, cube, mastar, '
                                   'image, rss, maps, modelcube, dap, default')

    assert binmode in [None, '*', 'MAPS', 'LOGCUBE'], 'binmode can only be *, MAPS or LOGCUBE'

    # Assert correct dir3d
    if dir3d != '*':
        assert dir3d in ['stack', 'mastar'], 'dir3d must be either stack or mastar'

    # Parse and retrieve the input type and the download type
    idtype = parseIdentifier(inputlist[0])
    if not idtype:
        raise MarvinError('Input list must be a list of plates, plate-ifus, or mangaids')

    # Set download type
    if dltype == 'cube':
        name = 'mangacube'
    elif dltype == 'rss':
        name = 'mangarss'
    elif dltype == 'default':
        name = 'mangadefault'
    elif dltype == 'plate':
        name = 'mangaplate'
    elif dltype == 'maps':
        # needs to change to include DR
        if '4' in release:
            name = 'mangamap'
        else:
            name = 'mangadap'
            binmode = 'MAPS'
    elif dltype == 'modelcube':
        name = 'mangadap'
        binmode = 'LOGCUBE'
    elif dltype == 'dap':
        name = 'mangadap'
        binmode = '*'
        daptype = '*'
    elif dltype == 'mastar':
        name = 'mangamastar'
    elif dltype == 'image':
        name = 'mangaimage'

    # create rsync
    rsync_access = Access(label='marvin_download', verbose=verbose, release=release)
    rsync_access.remote()

    # Add objects
    for item in inputlist:
        if idtype == 'mangaid':
            try:
                plateifu = mangaid2plateifu(item)
            except MarvinError:
                plateifu = None
            else:
                plateid, ifu = plateifu.split('-')
        elif idtype == 'plateifu':
            plateid, ifu = item.split('-')
        elif idtype == 'plate':
            plateid = item
            ifu = '*'

        rsync_access.add(name, plate=plateid, drpver=drpver, ifu=ifu, dapver=dapver, dir3d=dir3d,
                         mpl=release, bintype=bintype, n=n, mode=binmode, daptype=daptype,
                         wave=wave)

    # set the stream
    try:
        rsync_access.set_stream()
    except AccessError as e:
        raise MarvinError('Error with sdss_access rsync.set_stream. AccessError: {0}'.format(e))

    # get the list and download
    listofitems = rsync_access.get_urls() if as_url else rsync_access.get_paths()

    # print download location
    item = listofitems[0] if listofitems else None
    if item:
        ver = dapver if dapver in item else drpver
        dlpath = item[:item.rfind(ver) + len(ver)]
        if verbose:
            print('Target download directory: {0}'.format(dlpath))

    if test:
        return listofitems
    else:
        rsync_access.commit(limit=limit)


def _get_summary_file(name, summary_path=None, drpver=None, dapver=None):
    ''' Check for/download the drpall or dapall file

    Checks for existence of a local summary file for the
    current release set.  If not found, uses sdss_access
    to download it.

    Parameters:
        name (str):
            The name of the summary file.  Either drpall or dapall
        summary_path (str):
            The local path to either the drpall or dapall file
        drpver (str):
            The DRP version
        dapver (str):
            The DAP version
    '''

    assert name in ['drpall', 'dapall'], 'name must be either drpall or dapall'
    from marvin import config

    # # check for public release
    # is_public = 'DR' in config.release
    # release = config.release.lower() if is_public else None

    # get drpver and dapver
    config_drpver, config_dapver = config.lookUpVersions(config.release)
    drpver = drpver if drpver else config_drpver
    dapver = dapver if dapver else config_dapver

    if name == 'drpall' and not summary_path:
        summary_path = get_drpall_path(drpver)
    elif name == 'dapall' and not summary_path:
        summary_path = get_dapall_path(drpver, dapver)

    if not os.path.isfile(summary_path):
        warnings.warn('{0} file not found. Downloading it.'.format(name), MarvinUserWarning)
        rsync = Access(label='get_summary_file', release=config.release)
        rsync.remote()
        rsync.add(name, drpver=drpver, dapver=dapver)
        try:
            rsync.set_stream()
        except Exception as e:
            raise MarvinError('Could not download the {4} file with sdss_access: '
                              '{0}\nTry manually downloading it for version ({1},{2}) and '
                              'placing it {3}'.format(e, drpver, dapver, summary_path, name))
        else:
            rsync.commit()


def get_drpall_file(drpver=None, drpall=None):
    ''' Check for/download the drpall file

    Checks for existence of a local drpall file for the
    current release set.  If not found, uses sdss_access
    to download it.

    Parameters:
        drpver (str):
            The DRP version
        drpall (str):
            The local path to either the drpall file
    '''
    _get_summary_file('drpall', summary_path=drpall, drpver=drpver)


def get_dapall_file(drpver=None, dapver=None, dapall=None):
    ''' Check for/download the dapall file

    Checks for existence of a local dapall file for the
    current release set.  If not found, uses sdss_access
    to download it.

    Parameters:
        drpver (str):
            The DRP version
        dapver (str):
            The DAP version
        dapall (str):
            The local path to either the dapall file
    '''
    _get_summary_file('dapall', summary_path=dapall, drpver=drpver, dapver=dapver)


def get_drpall_row(plateifu, drpver=None, drpall=None):
    """Returns a dictionary from drpall matching the plateifu."""

    # get the drpall table
    drpall_table = get_drpall_table(drpver=drpver, drpall=drpall, hdu='MANGA')
    in_table = plateifu in drpall_table['plateifu']
    # check the mastar extension
    if not in_table:
        drpall_table = get_drpall_table(drpver=drpver, drpall=drpall, hdu='MASTAR')
        in_table = plateifu in drpall_table['plateifu']
        if not in_table:
            raise ValueError('No results found for {0} in drpall table'.format(plateifu))

    row = drpall_table[drpall_table['plateifu'] == plateifu]

    return row[0]


def _db_row_to_dict(row, remove_columns=False):
    """Converts a DB object to a dictionary."""

    from sqlalchemy.inspection import inspect as sa_inspect
    from sqlalchemy.ext.hybrid import hybrid_property

    row_dict = collections.OrderedDict()

    columns = row.__table__.columns.keys()

    mapper = sa_inspect(row.__class__)
    for key, item in mapper.all_orm_descriptors.items():

        if isinstance(item, hybrid_property):
            columns.append(key)

    for col in columns:
        if remove_columns and col in remove_columns:
            continue
        row_dict[col] = getattr(row, col)

    return row_dict


def get_nsa_data(mangaid, source='nsa', mode='auto', drpver=None, drpall=None):
    """Returns a dictionary of NSA data from the DB or from the drpall file.

    Parameters:
        mangaid (str):
            The mangaid of the target for which the NSA information will be returned.
        source ({'nsa', 'drpall'}):
            The data source. If ``source='nsa'``, the full NSA catalogue from the DB will
            be used. If ``source='drpall'``, the subset of NSA columns included in the drpall
            file will be returned.
        mode ({'auto', 'local', 'remote'}):
            See :ref:`mode-decision-tree`.
        drpver (str or None):
            The version of the DRP to use, if ``source='drpall'``. If ``None``, uses the
            version set by ``marvin.config.release``.
        drpall (str or None):
            A path to the drpall file to use if ``source='drpall'``. If not defined, the
            default drpall file matching ``drpver`` will be used.

    Returns:
        nsa_data (dict):
            A dictionary containing the columns and values from the NSA catalogue for
            ``mangaid``.

    """

    from marvin import config, marvindb
    from .structs import DotableCaseInsensitive

    valid_modes = ['auto', 'local', 'remote']
    assert mode in valid_modes, 'mode must be one of {0}'.format(valid_modes)

    valid_sources = ['nsa', 'drpall']
    assert source in valid_sources, 'source must be one of {0}'.format(valid_sources)

    log.debug('get_nsa_data: getting NSA data for mangaid=%r with source=%r, mode=%r',
              mangaid, source, mode)

    if mode == 'auto':
        log.debug('get_nsa_data: running auto mode mode.')
        try:
            nsa_data = get_nsa_data(mangaid, mode='local', source=source,
                                    drpver=drpver, drpall=drpall)
            return nsa_data
        except MarvinError as ee:
            log.debug('get_nsa_data: local mode failed with error %s', str(ee))
            try:
                nsa_data = get_nsa_data(mangaid, mode='remote', source=source,
                                        drpver=drpver, drpall=drpall)
                return nsa_data
            except MarvinError as ee:
                raise MarvinError('get_nsa_data: failed to get NSA data for mangaid=%r in '
                                  'auto mode with with error: %s', mangaid, str(ee))

    elif mode == 'local':

        if source == 'nsa':

            if config.db is not None:

                session = marvindb.session
                sampledb = marvindb.sampledb

                nsa_row = session.query(sampledb.NSA).join(sampledb.MangaTargetToNSA,
                                                           sampledb.MangaTarget).filter(
                    sampledb.MangaTarget.mangaid == mangaid).use_cache().all()

                if len(nsa_row) == 1:
                    return DotableCaseInsensitive(
                        _db_row_to_dict(nsa_row[0], remove_columns=['pk', 'catalogue_pk']))
                elif len(nsa_row) > 1:
                    warnings.warn('get_nsa_data: multiple NSA rows found for mangaid={0}. '
                                  'Using the first one.'.format(mangaid), MarvinUserWarning)
                    return DotableCaseInsensitive(
                        _db_row_to_dict(nsa_row[0], remove_columns=['pk', 'catalogue_pk']))
                elif len(nsa_row) == 0:
                    raise MarvinError('get_nsa_data: cannot find NSA row for mangaid={0}'
                                      .format(mangaid))

            else:

                raise MarvinError('get_nsa_data: cannot find a valid DB connection.')

        elif source == 'drpall':

            plateifu = mangaid2plateifu(mangaid, drpver=drpver, drpall=drpall, mode='drpall')
            log.debug('get_nsa_data: found plateifu=%r for mangaid=%r', plateifu, mangaid)

            drpall_row = get_drpall_row(plateifu, drpall=drpall, drpver=drpver)

            nsa_data = collections.OrderedDict()
            for col in drpall_row.colnames:
                if col.startswith('nsa_'):
                    value = drpall_row[col]
                    if isinstance(value, np.ndarray):
                        value = value.tolist()
                    else:
                        # In Astropy 2 the value would be an array of size 1
                        # but in Astropy 3 value is already an scalar and asscalar fails.
                        try:
                            value = np.asscalar(value)
                        except AttributeError:
                            pass
                    nsa_data[col[4:]] = value

            return DotableCaseInsensitive(nsa_data)

    elif mode == 'remote':

        from marvin.api.api import Interaction

        try:
            if source == 'nsa':
                request_name = 'nsa_full'
            else:
                request_name = 'nsa_drpall'
            url = marvin.config.urlmap['api'][request_name]['url']
            response = Interaction(url.format(mangaid=mangaid))
        except MarvinError as ee:
            raise MarvinError('API call to {0} failed: {1}'.format(request_name, str(ee)))
        else:
            if response.results['status'] == 1:
                return DotableCaseInsensitive(collections.OrderedDict(response.getData()))
            else:
                raise MarvinError('get_nsa_data: %s', response['error'])


def _check_file_parameters(obj1, obj2):
    for param in ['plateifu', 'mangaid', 'plate', '_release', 'drpver', 'dapver']:
        assert_msg = ('{0} is different between {1} {2}:\n {1}.{0}: {3} {2}.{0}:{4}'
                      .format(param, obj1.__repr__, obj2.__repr__, getattr(obj1, param),
                              getattr(obj2, param)))
        assert getattr(obj1, param) == getattr(obj2, param), assert_msg


def add_doc(value):
    """Wrap method to programatically add docstring."""
    def _doc(func):
        func.__doc__ = value
        return func
    return _doc


def use_inspect(func):
    ''' Inspect a function of arguments and keywords.

    Inspects a function or class method.  Uses a different inspect for Python 2 vs 3
    Only tested to work with args and defaults.  varargs (variable arguments)
    and varkw (keyword arguments) seem to always be empty.

    Parameters:
        func (func):
            The function or method to inspect

    Returns:
        A tuple of arguments, variable arguments, keywords, and default values

    '''
    pyver = sys.version_info.major
    if pyver == 2:
        args, varargs, varkw, defaults = inspect.getargspec(func)
    elif pyver == 3:
        sig = inspect.signature(func)
        args = []
        defaults = []
        varargs = varkw = None
        for par in sig.parameters.values():
            # most parameters seem to be of this kind
            if par.kind == par.POSITIONAL_OR_KEYWORD:
                args.append(par.name)
                # parameters with default of inspect empty are required
                if par.default != inspect._empty:
                    defaults.append(par.default)

    return args, varargs, varkw, defaults


def getRequiredArgs(func):
    ''' Gets the required arguments from a function or method

    Uses this difference between arguments and defaults to indicate
    required versus optional arguments

    Parameters:
        func (func):
            The function or method to inspect

    Returns:
        A list of required arguments

    Example:
        >>> import matplotlib.pyplot as plt
        >>> getRequiredArgs(plt.scatter)
        >>> ['x', 'y']

    '''
    args, varargs, varkw, defaults = use_inspect(func)
    if defaults:
        args = args[:-len(defaults)]
    return args


def getKeywordArgs(func):
    ''' Gets the keyword arguments from a function or method

    Parameters:
        func (func):
            The function or method to inspect

    Returns:
        A list of keyword arguments

    Example:
        >>> import matplotlib.pyplot as plt
        >>> getKeywordArgs(plt.scatter)
        >>> ['edgecolors', 'c', 'vmin', 'linewidths', 'marker', 's', 'cmap',
        >>>  'verts', 'vmax', 'alpha', 'hold', 'data', 'norm']

    '''
    args, varargs, varkw, defaults = use_inspect(func)
    req_args = getRequiredArgs(func)
    opt_args = list(set(args) - set(req_args))
    return opt_args


def missingArgs(func, argdict, arg_type='args'):
    ''' Return missing arguments from an input dictionary

    Parameters:
        func (func):
            The function or method to inspect
        argdict (dict):
            The argument dictionary to test against
        arg_type (str):
            The type of arguments to test.  Either (args|kwargs|req|opt). Default is required.

    Returns:
        A list of missing arguments

    Example:
        >>> import matplotlib.pyplot as plt
        >>> testdict = {'edgecolors': 'black', 'c': 'r', 'xlim': 5, 'xlabel': 9, 'ylabel': 'y', 'ylim': 6}
        >>> # test for missing required args
        >>> missginArgs(plt.scatter, testdict)
        >>> {'x', 'y'}
        >>> # test for missing optional args
        >>> missingArgs(plt.scatter, testdict, arg_type='opt')
        >>> ['vmin', 'linewidths', 'marker', 's', 'cmap', 'verts', 'vmax', 'alpha', 'hold', 'data', 'norm']

    '''
    assert arg_type in ['args', 'req', 'kwargs', 'opt'], 'arg_type must be one of (args|req|kwargs|opt)'
    if arg_type in ['args', 'req']:
        return set(getRequiredArgs(func)).difference(argdict)
    elif arg_type in ['kwargs', 'opt']:
        return set(getKeywordArgs(func)).difference(argdict)


def invalidArgs(func, argdict):
    ''' Return invalid arguments from an input dictionary

    Parameters:
        func (func):
            The function or method to inspect
        argdict (dict):
            The argument dictionary to test against

    Returns:
        A list of invalid arguments

    Example:
        >>> import matplotlib.pyplot as plt
        >>> testdict = {'edgecolors': 'black', 'c': 'r', 'xlim': 5, 'xlabel': 9, 'ylabel': 'y', 'ylim': 6}
        >>> # test for invalid args
        >>> invalidArgs(plt.scatter, testdict)
        >>>  {'xlabel', 'xlim', 'ylabel', 'ylim'}

    '''
    args, varargs, varkw, defaults = use_inspect(func)
    return set(argdict) - set(args)


def isCallableWithArgs(func, argdict, arg_type='opt', strict=False):
    ''' Test if the function is callable with the an input dictionary

    Parameters:
        func (func):
            The function or method to inspect
        argdict (dict):
            The argument dictionary to test against
        arg_type (str):
            The type of arguments to test.  Either (args|kwargs|req|opt). Default is required.
        strict (bool):
            If True, validates input dictionary against both missing and invalid keyword arguments. Default is False

    Returns:
        Boolean indicating whether the function is callable

    Example:
        >>> import matplotlib.pyplot as plt
        >>> testdict = {'edgecolors': 'black', 'c': 'r', 'xlim': 5, 'xlabel': 9, 'ylabel': 'y', 'ylim': 6}
        >>> # test for invalid args
        >>> isCallableWithArgs(plt.scatter, testdict)
        >>> False

    '''
    if strict:
        return not missingArgs(func, argdict, arg_type=arg_type) and not invalidArgs(func, argdict)
    else:
        return not invalidArgs(func, argdict)


def map_bins_to_column(column, indices):
    ''' Maps a dictionary of array indices to column data

    Takes a given data column and a dictionary of indices (see the indices key
    from output of the histgram data in :meth:`marvin.utils.plot.scatter.hist`),
    and produces a dictionary with the data values from column mapped in
    individual bins.

    Parameters:
        column (list):
            A column of data
        indices (dict):
            A dictionary of providing a list of array indices belonging to each
            bin in a histogram.

    Returns:
        A dictionary containing, for each binid, a list of column data in that bin.

    Example:
        >>>
        >>> # provide a list of data in each bin of an output histogram
        >>> x = np.random.random(10)*10
        >>> hdata = hist(x, bins=3, return_fig=False)
        >>> inds = hdata['indices']
        >>> pmap = map_bins_to_column(x, inds)
        >>> OrderedDict([(1,
        >>>   [2.5092488009906235,
        >>>    1.7494530589363955,
        >>>    2.5070840461208754,
        >>>    2.188355400587354,
        >>>    2.6987990403658992,
        >>>    1.6023553861428441]),
        >>>  (3, [7.9214280403215875, 7.488908995456573, 7.190598204420587]),
        >>>  (4, [8.533028236560906])])

    '''
    assert isinstance(indices, dict) is True, 'indices must be a dictionary of binids'
    assert len(column) == sum(map(len, indices.values())), 'input column and indices values must have same len'
    coldict = OrderedDict()
    colarr = np.array(column)
    for key, val in indices.items():
        coldict[key] = colarr[val].tolist()
    return coldict


def _sort_dir(instance, class_):
    """Sort `dir()` to return child class attributes and members first.

    Return the attributes and members of the child class, so that
    ipython tab completion lists those first.

    Parameters:
        instance: Instance of `class_` (usually self).
        class_: Class of `instance`.

    Returns:
        list: Child class attributes and members.
    """
    members_array = list(zip(*inspect.getmembers(np.ndarray)))[0]
    members_quantity = list(zip(*inspect.getmembers(Quantity)))[0]
    members_parents = members_array + members_quantity

    return_list = [it[0] for it in inspect.getmembers(class_) if it[0] not in members_parents]
    return_list += vars(instance).keys()
    return_list += ['value']
    return return_list


def _get_summary_path(name, drpver, dapver=None):
    ''' Return the path for either the DRP or DAP summary file

    Parameters:
        name (str):
            The name of the summary file, either drpall or dapall
        drpver (str):
            The DRP version
        dapver (str):
            The DAP version
    '''
    assert name in ['drpall', 'dapall'], 'name must be either drpall or dapall'
    release = marvin.config.lookUpRelease(drpver)
    # is_public = 'DR' in release
    # path_release = release.lower() if is_public else None
    path = Path(release=release)
    all_path = path.full(name, drpver=drpver, dapver=dapver)
    return all_path


def get_drpall_path(drpver):
    """Returns the path to the DRPall file for ``(drpver, dapver)``."""

    drpall_path = _get_summary_path('drpall', drpver=drpver)
    return drpall_path


def get_dapall_path(drpver, dapver):
    """Returns the path to the DAPall file for ``(drpver, dapver)``."""

    dapall_path = _get_summary_path('dapall', drpver, dapver)
    return dapall_path


@contextlib.contextmanager
def turn_off_ion(show_plot=True):
    ''' Turns off the Matplotlib plt interactive mode

    Context manager to temporarily disable the interactive
    Matplotlib plotting functionality.  Useful for only returning
    Figure and Axes objects

    Parameters:
        show_plot (bool):
            If True, turns off the plotting

    Example:
        >>>
        >>> with turn_off_ion(show_plot=False):
        >>>     do_some_stuff
        >>>

    '''

    plt_was_interactive = plt.isinteractive()
    if not show_plot and plt_was_interactive:
        plt.ioff()

    fignum_init = plt.get_fignums()

    yield plt

    if show_plot:
        plt.ioff()
        plt.show()
    else:
        for ii in plt.get_fignums():
            if ii not in fignum_init:
                plt.close(ii)

    # Restores original ion() status
    if plt_was_interactive and not plt.isinteractive():
        plt.ion()


@contextlib.contextmanager
def temp_setattr(ob, attrs, new_values):
    """ Temporarily set attributed on an object

    Temporarily set an attribute on an object for the duration of the
    context manager.

    Parameters:
        ob (object):
            A class instance to set attributes on
        attrs (str|list):
            A list of attribute names to replace
        new_values (list):
            A list of new values to set as new attribute.  If new_values is
            None, all attributes in attrs will be set to None.

    Example:
        >>> c = Cube(plateifu='8485-1901')
        >>> print('before', c.mangaid)
        >>> with temp_setattr(c, 'mangaid', None):
        >>>     # do stuff
        >>>     print('new', c.mangaid)
        >>> print('after' c.mangaid)
        >>>
        >>> # Output
        >>> before '1-209232'
        >>> new None
        >>> after '1-209232'
        >>>

    """

    # set up intial inputs
    attrs = attrs if isinstance(attrs, list) else [attrs]
    if new_values:
        new_values = new_values if isinstance(new_values, list) else [new_values]
    else:
        new_values = [new_values] * len(attrs)

    assert len(attrs) == len(new_values), 'attrs and new_values must have the same length'

    replaced = []
    old_values = []

    # grab the old values
    for i, attr in enumerate(attrs):
        new_value = new_values[i]

        replace = False
        old_value = None
        if hasattr(ob, attr):
            try:
                if attr in ob.__dict__:
                    replace = True
            except AttributeError:
                if attr in ob.__slots__:
                    replace = True
            if replace:
                old_value = getattr(ob, attr)
        replaced.append(replace)
        old_values.append(old_value)
        setattr(ob, attr, new_value)

    # yield
    yield replaced, old_values

    # replace the old values
    for i, attr in enumerate(attrs):
        if not replaced[i]:
            delattr(ob, attr)
        else:
            setattr(ob, attr, old_values[i])


def map_dapall(header, row):
    ''' Retrieves a dictionary of DAPall db column names

    For a given row in the DAPall file, returns a dictionary
    of corresponding DAPall database columns names with the
    appropriate values.

    Parameters:
        header (Astropy header):
            The primary header of the DAPall file
        row (recarray):
            A row of the DAPall binary table data

    Returns:
        A dictionary with db column names as keys and row data as values

    Example:
        >>> hdu = fits.open('dapall-v2_3_1-2.1.1.fits')
        >>> header = hdu[0].header
        >>> row = hdu[1].data[0]
        >>> dbdict = map_dapall(header, row)

    '''

    # get names from header
    emline_schannels = []
    emline_gchannels = []
    specindex_channels = []
    for key, val in header.items():
        if 'ELS' in key:
            emline_schannels.append(val.lower().replace('-', '_').replace('.', '_'))
        elif 'ELG' in key:
            emline_gchannels.append(val.lower().replace('-', '_').replace('.', '_'))
        elif re.search('SPI([0-9])', key):
            specindex_channels.append(val.lower().replace('-', '_').replace('.', '_'))

    # File column names
    names = row.array.names

    dbdict = {}
    for col in names:
        name = col.lower()
        shape = row[col].shape if hasattr(row[col], 'shape') else ()
        array = ''
        values = row[col]

        if len(shape) > 0:
            channels = shape[0]
            for i in range(channels):
                channame = emline_schannels[i] if 'emline_s' in name else \
                    emline_gchannels[i] if 'emline_g' in name else \
                    specindex_channels[i] if 'specindex' in name else i + 1
                colname = '{0}_{1}'.format(name, channame)
                dbdict[colname] = values[i]
        else:
            dbdict[name] = values

    return dbdict


def get_virtual_memory_usage_kb():
    """
    The process's current virtual memory size in Kb, as a float.

    Returns:
        A float of the virtual memory usage

    """

    assert psutil is not None, 'the psutil python package is required to run this function'

    return float(psutil.Process().memory_info().vms) / 1024.0


def memory_usage(where):
    """
    Print out a basic summary of memory usage.

    Parameters:
        where (str):
            A string description of where in the code you are summarizing memory usage
    """

    assert pympler is not None, 'the pympler python package is required to run this function'

    mem_summary = pympler.summary.summarize(pympler.muppy.get_objects())
    print("Memory summary: {0}".format(where))
    pympler.summary.print_(mem_summary, limit=2)
    print("VM: {0:.2f}Mb".format(get_virtual_memory_usage_kb() / 1024.0))


def target_status(mangaid, mode='auto', source='nsa', drpall=None, drpver=None):
    ''' Check the status of a MaNGA target

    Given a mangaid, checks the status of a target.  Will check if
    target exists in the NSA catalog (i.e. is a proper target) and checks if
    target has a corresponding plate-IFU designation (i.e. has been observed).

    Returns a string status indicating if a target has been observed, has not
    yet been observed, or is not a valid MaNGA target.

    Parameters:
        mangaid (str):
            The mangaid of the target to check for observed status
        mode ({'auto', 'drpall', 'db', 'remote', 'local'}):
            See mode in :func:`mangaid2plateifu` and :func:`get_nsa_data`.
        source ({'nsa', 'drpall'}):
            The NSA catalog data source. See source in :func:`get_nsa_data`.
        drpall (str or None):
            The drpall file to use.  See drpall in :func:`mangaid2plateifu` and :func:`get_nsa_data`.
        drpver (str or None):
            The DRP version to use.  See drpver in :func:`mangaid2plateifu` and :func:`get_nsa_data`.

    Returns:
        A status of "observed", "not yet observed", or "not valid target"

    '''

    # check for plateifu - target has been observed
    try:
        plateifu = mangaid2plateifu(mangaid, mode=mode, drpver=drpver, drpall=drpall)
    except (MarvinError, BrainError) as e:
        plateifu = None

    # check if target in NSA catalog - proper manga target
    try:
        nsa = get_nsa_data(mangaid, source=source, mode=mode, drpver=drpver, drpall=drpall)
    except (MarvinError, BrainError) as e:
        nsa = None

    # return observed boolean
    if plateifu and nsa:
        status = 'observed'
    elif not plateifu and nsa:
        status = 'not yet observed'
    elif not plateifu and not nsa:
        status = 'not valid target'

    return status


def target_is_observed(mangaid, mode='auto', source='nsa', drpall=None, drpver=None):
    ''' Check if a MaNGA target has been observed or not

    See :func:`target_status` for full documentation.

    Returns:
        True if the target has been observed.

    '''
    # check the target status
    status = target_status(mangaid, source=source, mode=mode, drpver=drpver, drpall=drpall)
    return status == 'observed'


def target_is_mastar(plateifu, drpver=None, drpall=None):
    ''' Check if a target is bright-time MaStar target

    Uses the local drpall file to check if a plateifu is a MaStar target

    Parameters:
        plateifu (str):
            The plateifu of the target
        drpver (str):
            The drpver version to check against
        drpall (str):
            The drpall file path

    Returns:
        True if it is

    '''

    row = get_drpall_row(plateifu, drpver=drpver, drpall=drpall)
    return row['srvymode'] == 'APOGEE lead'


def get_drpall_table(drpver=None, drpall=None, hdu='MANGA'):
    ''' Gets the drpall table

    Gets the drpall table either from cache or loads it. For releases
    of MPL-8 and up, galaxies are in the MANGA extension, and mastar
    targets are in the MASTAR extension, specified with the hdu keyword. For
    MPLs 1-7, there is only one data extension, which is read.

    Parameters:
        drpver (str):
            The DRP release version to load.  Defaults to current marvin release
        drpall (str):
            The full path to the drpall table. Defaults to current marvin release.
        hdu (str):
            The name of the HDU to read in.  Default is 'MANGA'

    Returns:
        An Astropy Table
    '''

    from marvin import config
    assert hdu.lower() in ['manga', 'mastar'], 'hdu can either be MANGA or MASTAR'
    hdu = hdu.upper()

    # get the drpall file
    get_drpall_file(drpall=drpall, drpver=drpver)

    # Loads the drpall table if it was not cached from a previous session.
    config_drpver, __ = config.lookUpVersions()
    drpver = drpver if drpver else config_drpver

    # check for drpver
    if drpver not in drpTable:
        drpTable[drpver] = {}

    # check for hdu
    hduext = hdu if check_versions(drpver, 'v2_5_3') else 'MANGA'
    if hdu not in drpTable[drpver]:
        drpall = drpall if drpall else get_drpall_path(drpver=drpver)
        data = {hduext: table.Table.read(drpall, hdu=hduext)}
        drpTable[drpver].update(data)

    drpall_table = drpTable[drpver][hduext]

    return drpall_table


def get_dapall_table(drpver=None, dapver=None, dapall=None):
    ''' Gets the dapall table

    Gets the dapall table either from cache or loads it. For releases
    of MPL-6 and up.

    Parameters:
        drpver (str):
            The DRP release version to load.  Defaults to current marvin release
        dapall (str):
            The full path to the dapall table. Defaults to current marvin release.

    Returns:
        An Astropy Table
    '''

    from marvin import config

    # get the dapall file
    get_dapall_file(dapall=dapall, drpver=drpver, dapver=dapver)

    # Loads the dapall table if it was not cached from a previous session.
    config_drpver, config_dapver = config.lookUpVersions(config.release)
    drpver = drpver if drpver else config_drpver
    dapver = dapver if dapver else config_dapver

    # check for dapver
    if dapver not in dapTable:
        dapall = dapall if dapall else get_dapall_path(drpver=drpver, dapver=dapver)
        data = table.Table.read(dapall, hdu=1)
        dapTable[dapver] = data

    dapall_table = dapTable[dapver]

    return dapall_table


def get_plates(drpver=None, drpall=None, release=None):
    ''' Get a list of unique plates from the drpall file

    Parameters:
        drpver (str):
            The DRP release version to load.  Defaults to current marvin release
        drpall (str):
            The full path to the drpall table. Defaults to current marvin release.
        release (str):
            The marvin release

    Returns:
        A list of plate ids

    '''
    assert not all([drpver, release]), 'Cannot set both drpver and release '

    if release:
        drpver, __ = marvin.config.lookUpVersions(release)

    drpall_table = get_drpall_table(drpver=drpver, drpall=drpall)
    plates = list(set(drpall_table['plate']))
    return plates


def check_versions(version1, version2):
    ''' Compate two version ids against each other

    Checks if version1 is greater than or equal to version2.

    Parameters:
        version1 (str):
            The version to check
        version2 (str):
            The version to check against

    Returns:
        A boolean indicating if version1 is >= version2
    '''

    return parse_version(version1) >= parse_version(version2)


def get_manga_image(cube=None, drpver=None, plate=None, ifu=None, dir3d=None, local=None, public=None):
    ''' Get a MaNGA IFU optical PNG image

    Parameters:
        cube (Cube):
            A Marvin Cube instance
        drpver (str):
            The drpver version
        plate (str|int):
            The plate id
        ifu (str|int):
            An IFU designation
        dir3d (str):
            The directory for 3d data, either 'stack' or 'mastar'
        local (bool):
            If True, return the local file path to the image
        public (bool):
            If True, use only DR releases
    Returns:
        A url to an IFU MaNGA image
    '''

    # check inputs
    drpver = cube._drpver if cube else drpver
    plate = cube.plate if cube else plate
    ifu = cube.ifu if cube else ifu
    dir3d = cube.dir3d if cube else dir3d
    assert all([drpver, plate, ifu]), 'drpver, plate, and ifu must be specified'

    # create the sdss Path
    release = cube.release if cube else marvin.config.lookUpRelease(drpver, public_only=public)
    path = Path(release=release)

    dir3d = dir3d if dir3d else 'stack'
    assert dir3d in ['stack', 'mastar'], 'dir3d can only be stack or mastar'
    if local:
        img = path.full('mangaimage', drpver=drpver, plate=plate, ifu=ifu, dir3d=dir3d)
    else:
        img = path.url('mangaimage', drpver=drpver, plate=plate, ifu=ifu, dir3d=dir3d)
    return img



#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-04-26 09:20:35
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-04-26 09:20:35 by Brian Cherinka
    Last Modified On: 2016-04-26 09:20:35 by Brian

'''
from __future__ import print_function
from __future__ import division
from brain.api.api import BrainInteraction
from marvin import config

configkeys = ['release', 'session_id', 'compression']


class Interaction(BrainInteraction):
    ''' Marvins Interaction class, subclassed from Brain

    This is the main class to make calls to the Marvin API.  Instantaiate
    the Interaction object with a URL to make the call.

    GET requests can be made without passing parameters.
    POST requests require parameters to be passed.

    A successful call results in a HTTP status code of 200.  Failures result
    in some other HTTP status code.  Results from the successful call are stored
    in an sttribute called results, as a dictionary.  Any data requested is
    stored as a key inside results called "data"

    Parameters:
        route (str):
            Required.  Relative url path of the API call you want to make
        params (dict):
            dictionary of parameters you are passing into the API function call
        request_type (str):
            the method type of the API call, can be either "get" or "post" (default: post)
        auth (str):
            the authentication method used for the API.  Currently set as default to use
            JSON Web token authentication.
        timeout (float|tuple):
            A float or tuple of floats indicating the request timeout limit in seconds.
            If the server has not sent a respsonse by the time limit, an exception is raised.
            The default timeout is set to 5 minutes.
            See http://docs.python-requests.org/en/master/user/advanced/#timeouts
        headers (dict):
            A custom header to send with the request
        stream (bool):
            If True, iterates over the response data.  Default is False.  When set, avoids reading the
            content all at once into memory for large responses.
            See `request streaming <http://docs.python-requests.org/en/master/user/advanced/#streaming-requests>`_
        datastream (bool):
            If True, expects the response content to be streamed back using a Python generator.
            All matters when Marvin Query return_all is True.
        base (str):
            Optional replacement for domain API url.

    Returns:
        results (dict):
            The **Response JSON object** from the API call.  If the API is successful, the json data is extracted
            from the response and stored in this dictionary.  See :ref:`marvin-api-routes` for a
            description of the contents of results in each route.

    Examples:
        >>> from marvin import config
        >>> config.mode = 'remote'
        >>>
        >>> # import the Marvin Interaction class
        >>> from marvin.api.api import Interaction
        >>>
        >>> # get and format an API url to retrieve basic Cube properties
        >>> plateifu = '7443-12701'
        >>> url = config.urlmap['api']['getCube']['url']
        >>>
        >>> # create and send the request, and retrieve a response
        >>> response = Interaction(url.format(name=plateifu))
        >>>
        >>> # check your response's status code
        >>> print(response.status_code)
        >>> 200
        >>>
        >>> # get the data in your response
        >>> data = response.getData()
        >>> print(data)
    '''

    def _loadConfigParams(self):
        """Load the local configuration into a parameters dictionary to be sent with the request"""

        if self.params:
            for k in configkeys:
                if k not in self.params or self.params[k] is None:
                    self.params[k] = config.__getattribute__(k)
        else:
            self.params = {k: config.__getattribute__(k) for k in configkeys}

    def setAuth(self, authtype=None):
        ''' Set the authorization '''

        release = self.params['release'] if self.params and 'release' in self.params else config.release

        # disable authentication for specific cases (collab use of a DR, public access, or o)
        usingdr = (config.access == 'collab' and 'DR' in release)
        onserver = (config.access == 'collab' and config.db == 'lore')
        if usingdr or config.access == 'public' or onserver:
            authtype = None
        else:
            assert authtype is not None, 'Must have an authorization type set for collab access to MPLs!'

        super(Interaction, self).setAuth(authtype=authtype)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: explore.py
# Project: controllers
# Author: Brian Cherinka
# Created: Tuesday, 7th July 2020 3:45:32 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Tuesday, 7th July 2020 3:45:32 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from flask import Blueprint, render_template, request, redirect, after_this_request
from flask import url_for, flash, current_app, session as current_session, jsonify
from flask_classful import route
from marvin.core.exceptions import MarvinError
from marvin.web.controllers import BaseWebView
from marvin.utils.datamodel.dap import datamodel
from marvin.utils.general import parseIdentifier
from werkzeug.utils import secure_filename
import os
import pandas as pd
from brain.api.base import processRequest

from marvin.api.base import arg_validate as av
from marvin.web.controllers.galaxy import buildMapDict
from marvin.tools import Cube

explore = Blueprint("explore_page", __name__)


def allowed_file(filename, extensions):
    ''' check if the filename has an allowed extension '''
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in extensions


def read_targets(filename):
    ''' read the list of targets from a filename '''
    try:
        df = pd.read_csv(filename)
    except Exception:
        targets = []
    else:
        # get the first column
        targets = df[df.columns[0]].astype(str).tolist()
    return targets


def check_targets(targets):
    ''' check the target string is either plateifu or mangaid '''
    return [t for t in targets if parseIdentifier(t) in ['plateifu', 'mangaid']]

class Explore(BaseWebView):
    route_base = '/explore/'

    def __init__(self):
        ''' Initialize the route '''
        super(Explore, self).__init__('marvin-explore')
        self.explore = self.base.copy()
        self.explore['n_targs'] = 0
        self.explore['targetlist'] = ''
        self.explore['uploaded'] = False
        self.explore['targets'] = None
        self.explore['mapchoice'] = 'emline_gflux:ha_6564'
        self.explore['maps'] = None
        self.explore['mapmsgs'] = None
        self.explore['mapsloading'] = False

    def before_request(self, *args, **kwargs):
        ''' Do these things before a request to any route '''
        super(Explore, self).before_request(*args, **kwargs)
        self.reset_dict(self.explore, exclude=['n_targs', 'mapchoice'])

    @route('/', methods=['GET', 'POST'])
    def index(self):

        dm = datamodel[self._dapver]
        daplist = [p.full(web=True) for p in dm.properties]

        self.explore['dapbintemps'] = dm.get_bintemps(db_only=True)
        self.explore['dapmaps'] = daplist

        self.explore['targetlist'] = current_session.get('targetlist', '')
        self.explore['n_targs'] = current_session.get('n_targs', 0)

        return render_template('explore.html', **self.explore)

    @route('/upload/', methods=['GET', 'POST'], endpoint='upload')
    def upload_file(self):
        home = 'explore_page.Explore:index'
        if request.method == 'GET':
            flash('Method not allowed', 'error')
            return redirect(url_for(home))

        # no file part
        if 'file' not in request.files:
            flash('No file uploaded', 'error')
            return redirect(url_for(home))

        # empty file
        fileobj = request.files['file']
        if fileobj.filename == '':
            flash('No selected file', 'error')
            return redirect(url_for(home))

        # bad extension
        extensions = current_app.config['ALLOWED_EXTENSIONS']
        if not allowed_file(fileobj.filename, extensions):
            flash('File must be one of extension: {0}'.format(list(extensions)), 'error')
            return redirect(url_for(home))

        # save the file
        filename = secure_filename(fileobj.filename)
        filepath = os.path.join(current_app.config['UPLOAD_FOLDER'], filename)
        fileobj.save(filepath)

        # delete the file after the request
        @after_this_request
        def remove_file(response):
            if os.path.exists(filepath):
                os.remove(filepath)
            return response

        # bad size
        filesize = os.path.getsize(filepath)
        if filesize > current_app.config['MAX_CONTENT_LENGTH']:
            flash('Uploaded file is too large', 'error')
            return redirect(url_for(home)), 413

        # read targets
        targets = read_targets(filepath)
        if not targets:
            flash('Could not read targets from file.  Check file format.', 'error')
            return redirect(url_for(home))
        else:
            orig_count = len(targets)
            targets = check_targets(targets)
            if not targets:
                flash('No valid targets found in format of plateifu or mangaid', 'error')
                return redirect(url_for(home))

            # check number of targets
            n_targets = len(targets)

            if n_targets < orig_count:
                flash(f'Removed {orig_count - n_targets} invalid targets from list', 'warning')

            msg = 'Targets successfully uploaded.'
            if n_targets > 100:
                targets = targets[:100]
                msg += ' Only a limit of 100 targets allowed.'
            msg += ' Now select a map below.'

            flash(msg, 'info')
            targetlist = '&#10;'.join(targets)

            dm = datamodel[self._dapver]
            daplist = [p.full(web=True) for p in dm.properties]
            self.explore['dapmaps'] = daplist
            self.explore['dapbintemps'] = dm.get_bintemps(db_only=True)
            self.explore['targets'] = targets
            self.explore['targetlist'] = targetlist
            self.explore['n_targs'] = len(targets)
            return render_template('explore.html', **self.explore)

    @route('/maps/', methods=['GET', 'POST'], endpoint='maps')
    def get_maps(self):
        stuff = processRequest(request)
        home = 'explore_page.Explore:index'

        dm = datamodel[self._dapver]
        daplist = [p.full(web=True) for p in dm.properties]

        self.explore['dapmaps'] = daplist
        self.explore['dapbintemps'] = dm.get_bintemps(db_only=True)

        targetlist = stuff.get('targetlist', '')
        mapchoice = stuff.get('mapchoice', '')
        btchoice = stuff.get('btchoice', '')

        if not targetlist:
            flash('Must provide a list of targets', 'error')
            return redirect(url_for(home))

        if not mapchoice or not btchoice:
            flash('Must select a map and bintype', 'error')
            return redirect(url_for(home))

        targets = targetlist.split('\r\n')
        self.explore['targetlist'] = targetlist
        self.explore['targets'] = targets
        self.explore['n_targs'] = len(targets)
        self.explore['mapchoice'] = mapchoice

        self.explore['maps'] = True
        self.explore['mapmsgs'] = 'go'
        self.explore['mapsloading'] = True

        return render_template('explore.html', **self.explore)
    
    @route('/webmap/', methods=['POST'], endpoint='webmap')
    def updateMaps(self):
        args = av.manual_parse(self, request, use_params='explore', required='target')
        #stuff = processRequest(request)
        mapchoice = args.get('mapchoice', '')
        btchoice = args.get('btchoice', '')
        target = args.get('target', '')

        output = {'maps': None, 'msg': None}

        try:
            cube = Cube(target, release=self._release)
        except MarvinError:
            mapdict = {'data': None, 'msg': 'Error',
                        'plotparams': None}
            mapmsg = 'No cube available'
            output['maps'] = mapdict
            output['msg'] = mapmsg
            return jsonify(result=output)

        hasbin = btchoice.split('-')[0] in cube.get_available_bintypes() if btchoice else None

        try:
            mapdict = buildMapDict(cube, [mapchoice], self._dapver, bintemp=btchoice)
            mapmsg = None
        except Exception as e:
            mapdict = {'data': None, 'msg': 'Error', 'plotparams': None}
            if hasbin:
                mapmsg = 'Error getting maps: {0}'.format(e)
            else:
                mapmsg = 'No maps available for selected bintype {0}. Try a different one.'.format(
                    btchoice)
            output['maps'] = mapdict
            output['msg'] = mapmsg
            return jsonify(result=output)
        else:
            self.explore['mapstatus'] = 1
            output['maps'] = mapdict[0]
            output['msg'] = mapmsg
            return jsonify(result=output)
        

Explore.register(explore)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2017-10-30
# @Filename: spectrum.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-29 17:48:19


from __future__ import absolute_import, division, print_function

import matplotlib.pyplot as plt
import numpy as np
from astropy.units import Angstrom, CompositeUnit, Quantity

from .base_quantity import QuantityMixIn


class Spectrum(Quantity, QuantityMixIn):
    """A class representing an spectrum with extra functionality.

    Parameters:
        value (array-like):
            The 1-D array contianing the spectrum.
        wavelength (array-like, optional):
            The wavelength solution for ``value``.
        scale (float, optional):
            The scale factor of the spectrum value.
        unit (astropy.unit.Unit, optional):
            The unit of the spectrum.
        wavelength_unit (astropy.unit.Unit, optional):
            The units of the wavelength solution. Defaults to Angstrom.
        ivar (array-like, optional):
            The inverse variance array for ``value``.
        std (array-like, optional):
            The standard deviation associated with ``value``.
        mask (array-like, optional):
            The mask array for ``value``.
        pixmask_flag (str):
            The maskbit flag to be used to convert from mask bits to labels
            (e.g., MANGA_DRP3PIXMASK).
        kwargs (dict):
            Keyword arguments to be passed to `~astropy.units.Quantity` when
            it is initialised.

    Returns:
        spectrum:
            An astropy Quantity-like object that contains the spectrum, as well
            as inverse variance, mask, and wavelength (itself a Quantity array).

    """

    def __new__(cls, flux, wavelength, scale=None, unit=Angstrom,
                wavelength_unit=Angstrom, ivar=None, std=None,
                mask=None, dtype=None, copy=True, pixmask_flag=None, **kwargs):

        flux = np.array(flux)

        # If the scale is defined, creates a new composite unit with the input scale.
        if scale is not None:
            unit = CompositeUnit(unit.scale * scale, unit.bases, unit.powers)

        obj = Quantity(flux, unit=unit, dtype=dtype, copy=copy)
        obj = obj.view(cls)
        obj._set_unit(unit)

        obj.ivar = np.array(ivar) if ivar is not None else None
        obj.mask = np.array(mask) if mask is not None else None

        if std is not None:
            assert ivar is None, 'std and ivar cannot be used at the same time.'
            obj._std = np.array(std)

        assert wavelength is not None, 'invalid wavelength'

        if isinstance(wavelength, Quantity):
            obj.wavelength = wavelength
        else:
            obj.wavelength = np.array(wavelength) * wavelength_unit

        obj.pixmask_flag = pixmask_flag

        return obj

    def __array_finalize__(self, obj):

        if obj is None:
            return

        self.ivar = getattr(obj, 'ivar', None)
        self._std = getattr(obj, '_std', None)
        self.mask = getattr(obj, 'mask', None)
        self.wavelength = getattr(obj, 'wavelength', None)

        self.pixmask_flag = getattr(obj, 'pixmask_flag', None)

        self._set_unit(getattr(obj, 'unit', None))

    def __getitem__(self, sl):

        new_obj = super(Spectrum, self).__getitem__(sl)

        if type(new_obj) is not type(self):
            new_obj = self._new_view(new_obj)

        new_obj._set_unit(self.unit)

        new_obj.ivar = self.ivar.__getitem__(sl) if self.ivar is not None else self.ivar
        new_obj._std = self._std.__getitem__(sl) if self._std is not None else self._std
        new_obj.mask = self.mask.__getitem__(sl) if self.mask is not None else self.mask
        new_obj.wavelength = self.wavelength.__getitem__(sl) \
            if self.wavelength is not None else self.wavelength

        return new_obj

    @property
    def std(self):
        """The standard deviation of the measurement."""

        return self.error

    def plot(self, xlim=None, ylim=None, show_std=True, use_mask=True,
             n_sigma=1, xlabel='Wavelength', ylabel='Flux', show_units=True,
             plt_style='seaborn-darkgrid', figure=None, return_figure=False,
             title=None, ytrim='positive', **kwargs):
        """Plots the spectrum.

        Displays the spectrum showing, optionally, the :math:`n\\sigma` region,
        and applying the mask.

        Parameters:
            xlim,ylim (tuple or None):
                The range to display for the x- and y-axis, respectively,
                defined as a tuple of two elements ``[xmin, xmax]``. If
                the range is ``None``, the range for the xaxis range will be
                set automatically by matploltib while the yaxis limits will be
                optimised to reject extreme data points.
            show_std (bool):
                If ``True``, the :math:`n\\sigma` range above and below the
                spectrum values will be shown as a shadowed area. ``n_sigma``
                determines how many sigmas will be plotted.
            use_mask (bool):
                If ``True``, the region in which the mask is set to
                ``DONOTUSE`` will be masked out in the plot.
            show_units (bool):
                If ``True``, the units will be added to the axis labels.
            n_sigma (float):
                The number of standard deviations that will be shown if
                ``show_std=True``.
            xlabel,ylabel (str or None):
                The axis labels to be passed to the plot. If not defined, the y
                axis will be labelled as ``Flux`` and the x axis as
                ``Wavelength``.
            ytrim (str):
                The default y-limit behavior when no ylim specified. Can be "positive",
                "percentile", or "minmax". "positive" imposes a lower y bound of 0.
                "percentile" performs a 10% sigma clipping on the data to compute the bounds.
                "minmax" uses the straight min/max of the data.  Default is "positive".
            title (str):
                The title of the plot
            plt_style (str):
                Matplotlib style sheet to use. Default is 'seaborn-darkgrid'.
            figure (`~matplotlib.figure.Figure` or None):
                The matplotlib `~matplotlib.figure.Figure` object from which
                the axes must be created. If ``figure=None``, a new figure will
                be created.
            return_figure (bool):
                If ``True``, the matplotlib `~matplotlib.figure.Figure` object
                used will be returned along with the axes object.
            kwargs (dict):
                Keyword arguments to be passed to `~matplotlib.axes.Axes.plot`.

        Returns:
            axes:
                The `~matplotlib.axes.Axes` object containing the plot
                representing the spectrum. If ``return_figure=True``, a tuple
                will be returned of the form ``(ax, fig)``, where ``fig`` is
                the associated `~matplotlib.figure.Figure`.

        Example:

          >>> ax = spectrum.plot(n_sigma=3)
          >>> ax.show()

        We can change the range of the axes after the object has been created.

          >>> ax.set_xlim(6500, 6600)
          >>> ax.show()

        """

        with plt.style.context(plt_style):
            fig = plt.figure() if figure is None else figure
            ax = fig.add_subplot(111)

        if self.mask is None:
            use_mask = False

        if use_mask and 'DONOTUSE' in self.pixmask.schema.label.tolist():
            donotuse_mask = self.pixmask.get_mask('DONOTUSE') > 0
            value = self.masked
            wave = np.ma.array(self.wavelength.value, mask=donotuse_mask)
            std = self.std.value if self.std else None
        else:
            value = self.value
            wave = self.wavelength.value
            std = self.std.value if self.std else None

        ax.plot(wave, value, **kwargs)

        if show_std and self.std is not None:
            assert n_sigma > 0, 'invalid n_sigma value.'
            ax.fill_between(wave,
                            value + n_sigma * std,
                            value - n_sigma * std,
                            facecolor='b', alpha=0.3)

        ax.set_xlim(xlim)

        if ylim is None:
            if ytrim == 'positive':
                ylim = (0, None)
            elif ytrim == 'percentile':
                # Uses percentiles to get the optimal limits
                ylim_0 = -np.percentile(-value[value < 0], 90)
                ylim_1 = np.percentile(value[value > 0], 99.5)
                ylim = (ylim_0, ylim_1)
            elif ytrim == 'minmax':
                # let matplotlib decide
                ylim = (np.min(value), np.max(value))

        ax.set_ylim(ylim)

        if xlabel is not None:
            if show_units and isinstance(self.wavelength, Quantity):
                xlabel = '$\\mathrm{{{}}}\\,[${}$]$'.format(
                    xlabel, self.wavelength.unit.to_string('latex'))
            ax.set_xlabel(xlabel)

        if ylabel is not None:
            if show_units:
                ylabel = '$\\mathrm{{{}}}\\,[${}$]$'.format(ylabel,
                                                            self.unit.to_string('latex_inline'))
            ax.set_ylabel(ylabel)

        if title:
            ax.set_title(title)

        if return_figure:
            return (ax, fig)
        else:
            return ax



#!/usr/bin/env python
# encoding: utf-8
#
# spaxel.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#     11 Apr 2016 J. Sánchez-Gallego
#       Initial version


# from __future__ import division
# from __future__ import print_function
#
# import numpy as np
#
# from flask_classful import route
# from flask import jsonify
#
# from marvin.tools.spaxel import Spaxel
# from marvin.api.base import BaseView, arg_validate as av
# from marvin.core.exceptions import MarvinError
# from marvin.utils.general import parseIdentifier
#
#
# def _getSpaxel(name, x, y, **kwargs):
#     """Retrieves a Marvin Spaxel object."""
#
#     spaxel = None
#     results = {}
#
#     # Pop the release to remove a duplicate input to Maps
#     release = kwargs.pop('release', None)
#
#     # parse name into either mangaid or plateifu
#     try:
#         idtype = parseIdentifier(name)
#     except Exception as e:
#         results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(e))
#         return spaxel, results
#
#     try:
#         if idtype == 'plateifu':
#             plateifu = name
#             mangaid = None
#         elif idtype == 'mangaid':
#             mangaid = name
#             plateifu = None
#         else:
#             raise MarvinError('invalid plateifu or mangaid: {0}'.format(idtype))
#
#         spaxel = Spaxel(x=x, y=y, mangaid=mangaid, plateifu=plateifu,
#                         release=release, **kwargs)
#         results['status'] = 1
#     except Exception as e:
#         results['error'] = 'Failed to retrieve Spaxels {0}: {1}'.format(name, str(e))
#
#     return spaxel, results
#
#
# class SpaxelView(BaseView):
#     """Class describing API calls related to Spaxels."""
#
#     route_base = '/spaxels/'
#
#     @route('/<name>/spectra/<x>/<y>/', methods=['GET', 'POST'], endpoint='getSpectrum')
#     @av.check_args()
#     def spectrum(self, args, name, x, y):
#         """Returns a dictionary with the DRP spectrum for a spaxel.
#
#         Loads a DRP Cube and uses getSpaxel to retrieve the ``(x,y)``
#         spaxel. Returns a dictionary with the spectrum for that spaxel.
#
#         .. :quickref: Spaxel; Get a spectrum from a specific spaxel from a DRP cube
#
#         :param name: The name of the object as plate-ifu or mangaid
#         :param x: The x coordinate of the spaxel (origin is ``lower``)
#         :param y: The y coordinate of the spaxel (origin is ``lower``)
#         :form inconfig: json of any incoming parameters
#         :resjson int status: status of response. 1 if good, -1 if bad.
#         :resjson string error: error message, null if None
#         :resjson json inconfig: json of incoming configuration
#         :resjson json utahconfig: json of outcoming configuration
#         :resjson string traceback: traceback of an error, null if None
#         :resjson json data: dictionary of returned data
#         :json list flux: the spectrum flux array
#         :json list ivar: the spectrum ivar array
#         :json list mask: the spectrum mask array
#         :json list wavelength: the spectrum wavelength array
#         :json list specres: the spectrum spectral resolution array
#         :resheader Content-Type: application/json
#         :statuscode 200: no error
#         :statuscode 422: invalid input parameters
#
#         **Example request**:
#
#         .. sourcecode:: http
#
#            GET /marvin/api/spaxels/8485-1901/spectra/10/10/ HTTP/1.1
#            Host: api.sdss.org
#            Accept: application/json, */*
#
#         **Example response**:
#
#         .. sourcecode:: http
#
#            HTTP/1.1 200 OK
#            Content-Type: application/json
#            {
#               "status": 1,
#               "error": null,
#               "inconfig": {"release": "MPL-5"},
#               "utahconfig": {"release": "MPL-5", "mode": "local"},
#               "traceback": null,
#               "data": {"flux": [-0.001416, 0.0099, 0.0144, ...],
#                     "ivar": [134.613, 133.393, 132.094, ...],
#                     "mask": [0, 0, 0, ...],
#                     "wavelength": [3621.6, 3622.43, ..., 10353.8],
#                     "specres": [1026.83, 1027.07, 1027.3]
#               }
#            }
#
#         """
#
#         # Pop any args we don't want going into Spaxel
#         args = self._pop_args(args, arglist=['name', 'x', 'y'])
#
#         spaxel, results = _getSpaxel(name, x, y, maps=False, modelcube=False, **args)
#
#         self.update_results(results)
#
#         if spaxel is not None:
#             self.results['data'] = {'flux': spaxel.spectrum.value.tolist(),
#                                     'ivar': spaxel.spectrum.ivar.tolist(),
#                                     'mask': spaxel.spectrum.mask.tolist(),
#                                     'wavelength': spaxel.spectrum.wavelength.value.tolist(),
#                                     'specres': spaxel.specres.tolist()}
#
#         return jsonify(self.results)
#
#     @route('/<name>/properties/<template>/<x>/<y>/',
#            methods=['GET', 'POST'], endpoint='getProperties')
#     @av.check_args()
#     def properties(self, args, name, x, y, template):
#         """Returns a dictionary with the DAP properties for a spaxel.
#
#         Loads a DAP Maps and uses getSpaxel to retrieve the ``(x,y)``
#         spaxel. Returns a dictionary with the properties for that spaxel.
#
#         .. :quickref: Spaxel; Get DAP properties from a specific spaxel from a DAP Maps
#
#         :param name: The name of the object as plate-ifu or mangaid
#         :param x: The x coordinate of the spaxel (origin is ``lower``)
#         :param y: The y coordinate of the spaxel (origin is ``lower``)
#         :param template: The template associated with this maps. If none, default is used.
#         :form inconfig: json of any incoming parameters
#         :resjson int status: status of response. 1 if good, -1 if bad.
#         :resjson string error: error message, null if None
#         :resjson json inconfig: json of incoming configuration
#         :resjson json utahconfig: json of outcoming configuration
#         :resjson string traceback: traceback of an error, null if None
#         :resjson json data: dictionary of returned data
#         :json dict properties: the DAP properties for this spaxel
#         :resheader Content-Type: application/json
#         :statuscode 200: no error
#         :statuscode 422: invalid input parameters
#
#         **Example request**:
#
#         .. sourcecode:: http
#
#            GET /marvin/api/spaxels/8485-1901/properties/GAU-MILESHC/10/10/ HTTP/1.1
#            Host: api.sdss.org
#            Accept: application/json, */*
#
#         **Example response**:
#
#         .. sourcecode:: http
#
#            HTTP/1.1 200 OK
#            Content-Type: application/json
#            {
#               "status": 1,
#               "error": null,
#               "inconfig": {"release": "MPL-5"},
#               "utahconfig": {"release": "MPL-5", "mode": "local"},
#               "traceback": null,
#               "data": {"properties": {
#                           "bin_area": {
#                             "channel": null,
#                             "description": "Area of each bin.",
#                             "ivar": null,
#                             "mask": null,
#                             "name": "bin_area",
#                             "unit": "arcsec^2",
#                             "value": 0.5
#                           },
#                           ...
#                        }
#               }
#            }
#
#         """
#
#         # Pop any args we don't want going into Spaxel
#         args = self._pop_args(args, arglist=['name', 'x', 'y'])
#
#         spaxel, results = _getSpaxel(name, x, y, cube=False, modelcube=False, **args)
#
#         self.update_results(results)
#
#         if spaxel is not None:
#             spaxel_properties = {}
#             for name in spaxel.properties:
#                 prop = spaxel.properties[name]
#                 spaxel_properties[name] = {}
#                 for key in ['name', 'value', 'ivar', 'mask', 'description']:
#                     propval = getattr(prop, key)
#                     if type(propval).__module__ == np.__name__:
#                         propval = np.asscalar(str(propval))
#                     spaxel_properties[name][key] = propval
#                 spaxel_properties[name]['channel'] = prop.channel.name if prop.channel else None
#                 spaxel_properties[name]['unit'] = prop.unit.to_string()
#
#             self.results['data'] = {'properties': spaxel_properties}
#
#         return jsonify(self.results)
#
#     @route('/<name>/models/<template>/<x>/<y>/',
#            methods=['GET', 'POST'], endpoint='getModels')
#     @av.check_args()
#     def getModels(self, args, name, x, y, template):
#         """Returns a dictionary with the models for a spaxel.
#
#         Loads a ModelCube and uses getSpaxel to retrieve the ``(x,y)``
#         spaxel. Returns a dictionary with the models for that spaxel.
#
#         .. :quickref: Spaxel; Get the models for a specific spaxel from a DAP ModelCube
#
#         :param name: The name of the object as plate-ifu or mangaid
#         :param x: The x coordinate of the spaxel (origin is ``lower``)
#         :param y: The y coordinate of the spaxel (origin is ``lower``)
#         :param template: The template associated with this maps. If none, default is used.
#         :form inconfig: json of any incoming parameters
#         :resjson int status: status of response. 1 if good, -1 if bad.
#         :resjson string error: error message, null if None
#         :resjson json inconfig: json of incoming configuration
#         :resjson json utahconfig: json of outcoming configuration
#         :resjson string traceback: traceback of an error, null if None
#         :resjson json data: dictionary of returned data
#         :json list flux_array: flux of the binned spectrum
#         :json list flux_ivar: ivar of the binned spectrum
#         :json list flux_mask: mask of the binned spectrum and model
#         :json list model_array: best fitting model spectra
#         :json list model_emline: model spectrum with only emission lines
#         :json list model_emline_base: model of constant baseline fitted beneath emission lines
#         :json list model_emline_mask: bitmask that applies only to emission-line modeling
#         :json string bintype: the spectrum spectral resolution array
#         :json string template: the spectrum spectral resolution array
#         :resheader Content-Type: application/json
#         :statuscode 200: no error
#         :statuscode 422: invalid input parameters
#
#         **Example request**:
#
#         .. sourcecode:: http
#
#            GET /marvin/api/spaxels/8485-1901/models/GAU-MILESHC/10/10/ HTTP/1.1
#            Host: api.sdss.org
#            Accept: application/json, */*
#
#         **Example response**:
#
#         .. sourcecode:: http
#
#            HTTP/1.1 200 OK
#            Content-Type: application/json
#            {
#               "status": 1,
#               "error": null,
#               "inconfig": {"release": "MPL-5"},
#               "utahconfig": {"release": "MPL-5", "mode": "local"},
#               "traceback": null,
#               "data": {"bintype": "SPX",
#                     "template": "GAU-MILESHC",
#                     "flux_array": [-0.001416, 0.0099, 0.0144, ...],
#                     "flux_ivar": [134.613, 133.393, 132.094, ...],
#                     "flux_mask": [32, 32, 32, ...],
#                     "model_array": [0, 0, 0, ...],
#                     "model_emline": [0, 0, 0, ...],
#                     "model_emline_base": [0, 0, 0, ...],
#                     "model_emline_mask": [128, 128, 128, ...],
#               }
#            }
#
#         """
#
#         # Pop any args we don't want going into Spaxel
#         args = self._pop_args(args, arglist=['name', 'x', 'y'])
#
#         spaxel, results = _getSpaxel(name, x, y, cube=False, maps=False, **args)
#
#         self.update_results(results)
#
#         if spaxel is not None:
#
#             self.results['data'] = {
#                 'flux_array': spaxel.model_flux.value.tolist(),
#                 'flux_ivar': spaxel.model_flux.ivar.tolist(),
#                 'flux_mask': spaxel.model_flux.mask.tolist(),
#                 'model_array': spaxel.model.value.tolist(),
#                 'model_emline': spaxel.emline.value.tolist(),
#                 'model_emline_base': spaxel.emline_base.value.tolist(),
#                 'model_emline_mask': spaxel.emline.mask.tolist(),
#                 'bintype': spaxel.bintype.name,
#                 'template': spaxel.template.name}
#
#         return jsonify(self.results)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-17 23:36:31
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-19 00:37:28

from __future__ import print_function, division, absolute_import

import copy as copy_mod
import os

import astropy.table as table
from astropy import units as u

from marvin.core.exceptions import MarvinError
from marvin.contrib.vacs.base import VACMixIn
from marvin.utils.datamodel import DataModelList
from marvin.utils.general.structs import FuzzyList


class VACDataModel(object):
    """A class representing a VAC datamodel """

    def __init__(self, release, aliases=[], vacs=[]):

        self.release = release
        self.aliases = aliases

        self.vacs = VACList(vacs, parent=self)

    def __repr__(self):

        return ('<VACDataModel release={0!r}, n_vacs={1}>'
                .format(self.release, len(self.vacs)))

    def copy(self):
        """Returns a copy of the datamodel."""

        return copy_mod.deepcopy(self)

    def __eq__(self, value):
        """Uses fuzzywuzzy to return the closest property match."""

        vac_names = [vac.name for vac in self.vacs]

        if value in vac_names:
            return self.vacs[vac_names.index(value)]

        try:
            vac_best_match = self.vacs[value]
        except ValueError:
            vac_best_match = None

        if vac_best_match is None:
            raise ValueError('too ambiguous input {!r}'.format(value))
        elif vac_best_match is not None:
            return vac_best_match

    def __contains__(self, value):

        try:
            match = self.__eq__(value)
            if match is None:
                return False
            else:
                return True
        except ValueError:
            return False

    def __getitem__(self, value):
        return self == value


class VACDataModelList(DataModelList):
    """A dictionary of DRP datamodels."""

    base = {'VACDataModel': VACDataModel}


class VACList(FuzzyList):
    ''' A fuzzy list of available VACs '''

    def __init__(self, the_list, parent=None):

        self.parent = parent

        super(VACList, self).__init__([])

        for item in the_list:
            self.append(item, copy=True)

    def mapper(self, value):
        """Helper method for the fuzzy list to match on the vac name."""

        return value.name

    def append(self, value, copy=True):
        """Appends with copy."""

        append_obj = value if copy is False else copy_mod.deepcopy(value)
        append_obj.parent = self.parent

        if issubclass(append_obj, VACMixIn):
            super(VACList, self).append(append_obj)
        else:
            raise ValueError('invalid vac of type {!r}'.format(type(append_obj)))

    def list_names(self):
        """Returns a list with the names of the vacs in this list."""

        return [item.name for item in self]

    def to_table(self, pprint=False, description=True, max_width=1000):
        """Returns an astropy table with all the vacs in this datamodel.

        Parameters:
            pprint (bool):
                Whether the table should be printed to screen using astropy's
                table pretty print.
            description (bool):
                If ``True``, an extra column with the description of the
                vac will be added.
            max_width (int or None):
                A keyword to pass to ``astropy.table.Table.pprint()`` with the
                maximum width of the table, in characters.

        Returns:
            result (``astropy.table.Table``):
                If ``pprint=False``, returns an astropy table containing
                the name of the vac, whether it has ``ivar`` or
                ``mask``, the units, and a description (if
                ``description=True``)..

        """

        vac_table = table.Table(
            None, names=['name', 'version', 'description'],
            dtype=['S20', 'S20', 'S500'])

        if self.parent:
            vac_table.meta['release'] = self.parent.release

        for vac in self:
            vac_table.add_row((vac.name,
                               vac.version[self.parent.release],
                               vac.description))

        if not description:
            vac_table.remove_column('description')

        if pprint:
            vac_table.pprint(max_width=max_width, max_lines=1e6)
            return

        return vac_table

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        ''' Write the datamodel to a CSV '''

        release = self.parent.release.lower().replace('-', '')

        if not filename:
            filename = 'vacs_dm_{0}.csv'.format(release)

        if not path:
            path = os.path.join(os.getenv("MARVIN_DIR"), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)

